# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Data Model class creation and other utils.

Check :mod:`eve.datamodels` for additional information.
"""

from __future__ import annotations

import dataclasses
import functools
import sys
import typing
import warnings

import attr
import attrs


try:
    # For perfomance reasons, try to use cytoolz when possible (using cython)
    import cytoolz as toolz  # type: ignore[import]
except ModuleNotFoundError:
    # Fall back to pure Python toolz
    import toolz  # type: ignore[import] # noqa: F401

from .. import exceptions
from .. import extended_typing as xtyping
from .. import type_validation as type_val
from .. import utils
from ..extended_typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Final,
    ForwardRef,
    Generator,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeAlias,
    TypeAnnotation,
    TypeVar,
    Union,
    cast,
    overload,
)
from ..type_definitions import NOTHING, NothingType


# Typing
# TODO(egparedes): these typing definitions are not perfect, but they provide
#   some help until more advanced features are added to typing, like
#   PEP 681 - Data Class Transforms (https://peps.python.org/pep-0681/)
#   or intersection types.
_T = TypeVar("_T")


# TODO(egparedes): since these protocols are used instead of the actual classes
#   for type checking, we assign empty tuples and None values to its members
#   to avoid errors from mypy complaining about instantiation of abstract classes


class _AttrsClassTP(Protocol):
    __attrs_attrs__: ClassVar[Tuple[attr.Attribute, ...]] = ()


Attribute: TypeAlias = attr.Attribute


class DataModelTP(_AttrsClassTP, xtyping.DevToolsPrettyPrintable, Protocol):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    __datamodel_fields__: ClassVar[utils.FrozenNamespace[Attribute]] = cast(
        utils.FrozenNamespace[Attribute], None
    )
    __datamodel_params__: ClassVar[utils.FrozenNamespace[Any]] = cast(
        utils.FrozenNamespace[Attribute], None
    )
    __datamodel_root_validators__: ClassVar[
        Tuple[xtyping.NonDataDescriptor["DataModelTP", "BoundRootValidator"], ...]
    ] = ()
    # Optional
    __auto_init__: ClassVar[Callable[..., None]] = cast(Callable[..., None], None)
    __pre_init__: ClassVar[Callable[[DataModelTP], None]] = cast(
        Callable[["DataModelTP"], None], None
    )
    __post_init__: ClassVar[Callable[[DataModelTP], None]] = cast(
        Callable[["DataModelTP"], None], None
    )


DataModelT = TypeVar("DataModelT", bound=DataModelTP)


class GenericDataModelTP(DataModelTP, Protocol):
    __args__: ClassVar[Tuple[Union[Type, TypeVar], ...]] = ()
    __parameters__: ClassVar[Tuple[TypeVar, ...]] = ()

    @classmethod
    def __class_getitem__(
        cls: Type[GenericDataModelTP], args: Union[Type, Tuple[Type, ...]]
    ) -> Union[DataModelTP, GenericDataModelTP]:
        ...


_DM = TypeVar("_DM", bound="DataModel")

GenericDataModelT = TypeVar("GenericDataModelT", bound=GenericDataModelTP)

AttrsValidator = Callable[[Any, Attribute, _T], Any]
FieldValidator = Callable[[_DM, Attribute, _T], None]
BoundFieldValidator = Callable[[Attribute, _T], None]

RootValidator = Callable[[Type[_DM], _DM], None]
BoundRootValidator = Callable[[_DM], None]

FieldTypeValidatorFactory = Callable[[TypeAnnotation, str], FieldValidator]

TypeConverter = Callable[[Any], _T]

# Implementation
_DATAMODEL_TAG: Final = "__DATAMODEL_TAG"
_FIELD_VALIDATOR_TAG: Final = "__DATAMODEL_FIELD_VALIDATOR_TAG"
_ROOT_VALIDATOR_TAG: Final = "__DATAMODEL_ROOT_VALIDATOR_TAG"
_COERCED_TYPE_TAG: Final = "__DATAMODEL_COERCED_TYPE_TAG"
_UNCHECKED_TYPE_TAG: Final = "__DATAMODEL_UNCHECKED_TYPE_TAG"

_DM_OPTS = "__dm_opts"
_GENERIC_DATAMODEL_ROOT_DM_OPT: Final = "_GENERIC_DATAMODEL_ROOT_DM_OPT"

MODEL_FIELD_DEFINITIONS_ATTR: Final = "__datamodel_fields__"
MODEL_PARAM_DEFINITIONS_ATTR: Final = "__datamodel_params__"
MODEL_ROOT_VALIDATORS_ATTR: Final = "__datamodel_root_validators__"


Coerced = xtyping.Annotated[_T, _COERCED_TYPE_TAG]
"""Type hint marker to define fields that should be coerced at initialization."""


Unchecked = xtyping.Annotated[_T, _UNCHECKED_TYPE_TAG]
"""Type hint marker to define fields that should NOT be type-checked at initialization."""


if sys.version_info >= (3, 10):
    _dataclass_opts: Final[dict[str, Any]] = {"slots": True}
else:
    _dataclass_opts: Final[Dict[str, Any]] = {}


@dataclasses.dataclass(**_dataclass_opts)
class ForwardRefValidator:
    """Implementation of ``attrs`` field validator for ``ForwardRef`` typings.

    The first time is called it will update the class' field type annotations
    and then create the actual type validator for the field.
    """

    factory: type_val.TypeValidatorFactory
    """Type factory used to create the actual field validator."""

    validator: Union[type_val.FixedTypeValidator, None, NothingType] = NOTHING
    """Actual type validator created after resolving the forward references."""

    def __call__(self, instance: DataModel, attribute: Attribute, value: Any) -> None:
        if self.validator is NOTHING:
            model_cls = instance.__class__
            update_forward_refs(model_cls)
            self.validator = self.factory(
                getattr(getattr(model_cls, MODEL_FIELD_DEFINITIONS_ATTR), attribute.name).type,
                attribute.name,
            )

        if self.validator:
            self.validator(value)


@dataclasses.dataclass(frozen=True, **_dataclass_opts)
class ValidatorAdapter:
    """Adapter to use :class:`eve.type_validation.FixedTypeValidator`s as field validators."""

    validator: type_val.FixedTypeValidator
    description: str

    def __call__(self, _instance: DataModel, _attribute: Attribute, value: Any) -> None:
        self.validator(value)

    def __repr__(self) -> str:
        return self.description


def field_type_validator_factory(
    factory: type_val.TypeValidatorFactory, *, use_cache: bool = False
) -> FieldTypeValidatorFactory:
    """Create a factory of field type validators from a factory of regular type validators."""
    if use_cache:
        factory = cast(
            type_val.TypeValidatorFactory,
            utils.optional_lru_cache(func=factory),
        )

    def _field_type_validator_factory(
        type_annotation: TypeAnnotation,
        name: str,
    ) -> FieldValidator:
        """Field type validator for datamodels, supporting forward references."""
        if isinstance(type_annotation, ForwardRef):
            return ForwardRefValidator(factory)
        else:
            simple_validator = factory(type_annotation, name, required=True)
            return ValidatorAdapter(
                simple_validator, f"{getattr(simple_validator,'__name__', 'TypeValidator')}"
            )

    return _field_type_validator_factory


simple_type_validator_factory = field_type_validator_factory(type_val.simple_type_validator_factory)

DefaultFieldTypeValidatorFactory: Final[Optional[FieldTypeValidatorFactory]] = (
    simple_type_validator_factory if __debug__ else None
)
"""Default type validator factory used by datamodels classes. `None` by default if running in optimized mode."""

_REPR_DEFAULT: Final = True
_EQ_DEFAULT: Final = True
_ORDER_DEFAULT: Final = False
_UNSAFE_HASH_DEFAULT: Final = False
_FROZEN_DEFAULT: Final = False
_MATCH_ARGS_DEFAULT: Final = True
_KW_ONLY_DEFAULT: Final = False
_SLOTS_DEFAULT: Final = False
_COERCE_DEFAULT: Final = False
_GENERIC_DEFAULT: Final = False

DEFAULT_OPTIONS: Final[utils.FrozenNamespace] = utils.FrozenNamespace(
    repr=_REPR_DEFAULT,
    eq=_EQ_DEFAULT,
    order=_ORDER_DEFAULT,
    unsafe_hash=_UNSAFE_HASH_DEFAULT,
    frozen=_FROZEN_DEFAULT,
    match_args=_MATCH_ARGS_DEFAULT,
    kw_only=_KW_ONLY_DEFAULT,
    slots=_SLOTS_DEFAULT,
    coerce=_COERCE_DEFAULT,
    generic=_GENERIC_DEFAULT,
)
"""Convenient public namespace to expose default values to users."""


@overload
def datamodel(
    cls: Literal[None] = None,
    /,
    *,
    repr: bool = _REPR_DEFAULT,  # noqa: A002  # shadowing 'repr' python builtin
    eq: bool = _EQ_DEFAULT,
    order: bool = _ORDER_DEFAULT,
    unsafe_hash: bool = _UNSAFE_HASH_DEFAULT,
    frozen: bool | Literal["strict"] = _FROZEN_DEFAULT,
    match_args: bool = _MATCH_ARGS_DEFAULT,
    kw_only: bool = _KW_ONLY_DEFAULT,
    slots: bool = _SLOTS_DEFAULT,
    coerce: bool = _COERCE_DEFAULT,
    generic: bool = _GENERIC_DEFAULT,
    type_validation_factory: Optional[FieldTypeValidatorFactory] = DefaultFieldTypeValidatorFactory,
) -> Callable[[Type[_T]], Type[_T]]:
    ...


@overload
def datamodel(  # noqa: F811  # redefinion of unused symbol
    cls: Type[_T],
    /,
    *,
    repr: bool = _REPR_DEFAULT,  # noqa: A002  # shadowing 'repr' python builtin
    eq: bool = _EQ_DEFAULT,
    order: bool = _ORDER_DEFAULT,
    unsafe_hash: bool = _UNSAFE_HASH_DEFAULT,
    frozen: bool | Literal["strict"] = _FROZEN_DEFAULT,
    match_args: bool = _MATCH_ARGS_DEFAULT,
    kw_only: bool = _KW_ONLY_DEFAULT,
    slots: bool = _SLOTS_DEFAULT,
    coerce: bool = _COERCE_DEFAULT,
    generic: bool = _GENERIC_DEFAULT,
    type_validation_factory: Optional[FieldTypeValidatorFactory] = DefaultFieldTypeValidatorFactory,
) -> Type[_T]:
    ...


# TODO(egparedes): Use @dataclass_transform(eq_default=True, field_specifiers=("field",))
def datamodel(  # noqa: F811  # redefinion of unused symbol
    cls: Optional[Type[_T]] = None,
    /,
    *,
    repr: bool = _REPR_DEFAULT,  # noqa: A002  # shadowing 'repr' python builtin
    eq: bool = _EQ_DEFAULT,
    order: bool = _ORDER_DEFAULT,
    unsafe_hash: bool = _UNSAFE_HASH_DEFAULT,
    frozen: bool | Literal["strict"] = _FROZEN_DEFAULT,
    match_args: bool = _MATCH_ARGS_DEFAULT,
    kw_only: bool = _KW_ONLY_DEFAULT,
    slots: bool = _SLOTS_DEFAULT,
    coerce: bool = _COERCE_DEFAULT,
    generic: bool = _GENERIC_DEFAULT,
    type_validation_factory: Optional[FieldTypeValidatorFactory] = DefaultFieldTypeValidatorFactory,
) -> Union[Type[_T], Callable[[Type[_T]], Type[_T]]]:
    """Add generated special methods to classes according to the specified attributes (class decorator).

    It converts the class to an `attrs <https://www.attrs.org/>`_ with some extra features.
    Adding strict type validation functions for the fields is done by means of
    the ``type_validation_factory`` argument, falling back to the default factory
        (:class:`DefaultFieldTypeValidatorFactory`). The generated field type
    validators are generated by using PEP 526 ``__annotations__`` to determine field
    types and creating validation functions for them.

    Arguments:
        cls: Original class definition.

    Keyword Arguments:
        repr: If ``True`` (default), a ``__repr__()`` method will be generated if it does
            not overwrite a custom implementation defined in this class (not inherited).
        eq: If ``True`` (default), ``__eq__()`` and ``__ne__()`` methods will be generated
            if they do not overwrite custom implementations defined in this class (not inherited).
            The generated method compares the class as if it were a tuple of its fields, but both
            instances in the comparison must be of identical type.
        order:  If ``True`` (default is ``False``), add ``__lt__()``, ``__le__()``, ``__gt__()``,
            and ``__ge__()`` methods that behave like `eq` above and allow instances
            to be ordered. If ``None`` mirror value of `eq`.
        unsafe_hash: If ``False``, a ``__hash__()`` method is generated in a safe way
            according to how ``eq`` and ``frozen`` are set, or set to ``None`` (disabled)
            otherwise. If ``True``, a ``__hash__()`` method is generated anyway
            (use with care). See :func:`dataclasses.dataclass` for the complete explanation
            (or other sources like: `<https://hynek.me/articles/hashes-and-equality/>`_).
        frozen: If ``True`` (default is ``False``), assigning to fields will generate an exception.
            This emulates read-only frozen instances. The ``__setattr__()`` and
            ``__delattr__()`` methods should not be defined in the class.
        match_args: If ``True`` (default) and ``__match_args__`` is not already defined in the class,
            set ``__match_args__`` on the class to support PEP 634 (Structural Pattern Matching).
            It is a tuple of all positional-only ``__init__`` parameter names on
            Python 3.10 and later. Ignored on older Python versions.
        kw_only: If ``True`` (default is ``False``), make all fields keyword-only in the generated
            ``__init__`` (if ``init`` is ``False``, this parameter is ignored).
        slots: slots: If ``True`` (the default is ``False``), ``__slots__`` attribute will be generated
            and a new slotted class will be returned instead of the original one.
        coerce: If ``True`` (default is ``False``), an automatic type converter will be generated
            for all fields.
        generic: If ``True`` (default is ``False``) the class should be a ``Generic[]`` class,
            and the generated Data Model will support creation of new Data Model subclasses
            when using concrete types as arguments of ``DataModelClass[some_concret_type]``.
        type_validation_factory: Type validation factory used to build the field type validators.
            If ``None``, type validators will not be generators.

    """
    datamodel_options: Final = {
        "repr": repr,
        "eq": eq,
        "order": order,
        "unsafe_hash": unsafe_hash,
        "frozen": frozen,
        "match_args": match_args,
        "kw_only": kw_only,
        "slots": slots,
        "coerce": coerce,
        "generic": generic,
        "type_validation_factory": type_validation_factory,
    }

    if cls is None:  # called as: @datamodel()
        return functools.partial(_make_datamodel, **datamodel_options)
    else:  # called as: @datamodel
        return _make_datamodel(
            cls,
            _stacklevel_offset=1,
            **datamodel_options,  # type: ignore[arg-type]
        )


class _DataModelDecoratorTP(Protocol[_T]):
    def __call__(
        self,
        cls: Optional[Type[_T]] = None,
        /,
        *,
        repr: bool = _REPR_DEFAULT,  # noqa: A002  # shadowing 'repr' python builtin
        eq: bool = _EQ_DEFAULT,
        order: bool = _ORDER_DEFAULT,
        unsafe_hash: bool = _UNSAFE_HASH_DEFAULT,
        match_args: bool = _MATCH_ARGS_DEFAULT,
        kw_only: bool = _KW_ONLY_DEFAULT,
        slots: bool = _SLOTS_DEFAULT,
        coerce: bool = _COERCE_DEFAULT,
        generic: bool = _GENERIC_DEFAULT,
        type_validation_factory: Optional[
            FieldTypeValidatorFactory
        ] = DefaultFieldTypeValidatorFactory,
    ) -> Union[Type[_T], Callable[[Type[_T]], Type[_T]]]:
        ...


frozenmodel: _DataModelDecoratorTP = functools.partial(datamodel, frozen=True)
"""Data Model definition function using ``frozen=True``."""

frozen_model = frozenmodel


# Typing protocols are used instead of the actual classes for type checks
if xtyping.TYPE_CHECKING:

    class DataModel(DataModelTP):
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            ...

        def __pretty__(
            self, fmt: Callable[[Any], Any], **kwargs: Any
        ) -> Generator[Any, None, None]:
            ...

else:

    # TODO(egparedes): use @dataclass_transform(eq_default=True, field_specifiers=("field",))
    class DataModel:
        """Base class to automatically convert any subclass into a Data Model.

        Inheriting from this class is equivalent to apply the :func:`datamodel`
        decorator to a class, except that the ``slots`` option is always ``False``
        (since it generates a new class) and all descendants will be also converted
        automatically in Data Models with the same options of the parent class
        (which does not happen when explicitly applying the decorator).

        See :func:`datamodel` for the description of the parameters.
        """

        __slots__ = ()

        @classmethod
        def __init_subclass__(
            cls,
            /,
            *,
            repr: bool  # noqa: A002  # shadowing 'repr' python builtin
            | None
            | Literal["inherited"] = "inherited",
            eq: bool | None | Literal["inherited"] = "inherited",
            order: bool | None | Literal["inherited"] = "inherited",
            unsafe_hash: bool | None | Literal["inherited"] = "inherited",
            frozen: bool | Literal["strict", "inherited"] = "inherited",
            match_args: bool | Literal["inherited"] = "inherited",
            kw_only: bool | Literal["inherited"] = "inherited",
            coerce: bool | Literal["inherited"] = "inherited",
            type_validation_factory: Optional[FieldTypeValidatorFactory]
            | Literal["inherited"] = "inherited",
            **kwargs: Any,
        ) -> None:
            dm_opts = kwargs.pop(_DM_OPTS, [])
            super(DataModel, cls).__init_subclass__(**kwargs)
            cls_params = getattr(cls, MODEL_PARAM_DEFINITIONS_ATTR, None)

            generic: Final = (
                "True_no_checks"
                if _GENERIC_DATAMODEL_ROOT_DM_OPT in dm_opts
                else getattr(cls_params, "generic", False)
            )

            locals_ = locals()
            datamodel_kwargs = {}
            for arg_name, default_value in [
                ("repr", _REPR_DEFAULT),
                ("eq", _EQ_DEFAULT),
                ("order", _ORDER_DEFAULT),
                ("unsafe_hash", _UNSAFE_HASH_DEFAULT),
                ("frozen", _FROZEN_DEFAULT),
                ("match_args", _MATCH_ARGS_DEFAULT),
                ("kw_only", _KW_ONLY_DEFAULT),
                ("coerce", _COERCE_DEFAULT),
                ("type_validation_factory", DefaultFieldTypeValidatorFactory),
            ]:
                arg_value = locals_[arg_name]
                if arg_value == "inherited":
                    datamodel_kwargs[arg_name] = getattr(cls_params, arg_name, default_value)
                else:
                    datamodel_kwargs[arg_name] = arg_value

            if cls_params is not None and cls_params.frozen and not datamodel_kwargs["frozen"]:
                raise TypeError("Subclasses of a frozen DataModel cannot be unfrozen.")

            _make_datamodel(
                cls,
                slots=False,
                generic=generic,
                **datamodel_kwargs,
                _stacklevel_offset=1,
            )


def field(
    *,
    default: Any = NOTHING,
    default_factory: Optional[Callable[[], Any]] = None,
    init: bool = True,
    repr: bool = True,  # noqa: A002   # shadowing 'repr' python builtin
    hash: Optional[bool] = None,  # noqa: A002   # shadowing 'hash' python builtin
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    kw_only: bool = _KW_ONLY_DEFAULT,
    converter: Callable[[Any], Any] | Literal["coerce"] | None = None,
    validator: AttrsValidator
    | FieldValidator
    | Sequence[AttrsValidator | FieldValidator]
    | None = None,
) -> Any:  # attr.s lies in some typings
    """Define a new attribute on a class with advanced options.

    Keyword Arguments:
        default: If provided, this will be the default value for this field.
            This is needed because the ``field()`` call itself replaces the
            normal position of the default value.
        default_factory: If provided, it must be a zero-argument callable that will
            be called when a default value is needed for this field. Among other
            purposes, this can be used to specify fields with mutable default values.
            It is an error to specify both `default` and `default_factory`.
        init: If ``True`` (default), this field is included as a parameter to the
            generated ``__init__()`` method.
        repr: If ``True`` (default), this field is included in the string returned
            by the generated ``__repr__()`` method.
        hash: This can be a ``bool`` or ``None`` (default). If ``True``, this field is
            included in the generated ``__hash__()`` method. If ``None``, use the value
            of `compare`, which would normally be the expected behavior: a field
            should be considered in the `hash` if it is used for comparisons.
            Setting this value to anything other than ``None`` is `discouraged`.
        compare: If ``True`` (default), this field is included in the generated equality and
            comparison methods (__eq__(), __gt__(), et al.).
        metadata: An arbitrary mapping, not used at all by Data Models, and provided
            only as a third-party extension mechanism. Multiple third-parties can each
            have their own key, to use as a namespace in the metadata.
        kw_only: If ``True`` (default is ``False``), make this field keyword-only in the
            generated ``__init__`` (if ``init`` is ``False``, this parameter is ignored).
        converter: Callable that is automatically called to convert attribute’s value.
            It is given the passed-in value, and the returned value will be used as the
            new value of the attribute before being passed to the validator, if any.
            If ``"coerce"`` is passed, a naive coercer converter will be generated.
            The automatic converter basically calls the constructor of the type indicated
            in the type hint, so it is assumed that new instances of this type can be created
            like ``type_name(value)``. For collection types, there is not attempt to convert
            its items, only the collection type.
        validator: FieldValidator or list of FieldValidators to be used with this field.
            (Note that validators can also be set using decorator notation).


    Examples:
        >>> from typing import List
        >>> @datamodel
        ... class C:
        ...     mylist: List[int] = field(default_factory=lambda : [1, 2, 3])
        >>> c = C()
        >>> c.mylist
        [1, 2, 3]

    """
    if default is not NOTHING and default_factory is not None:
        raise ValueError("Cannot specify both 'default' and 'default_factory'.")

    if default is not NOTHING:
        defaults_kwargs = {"default": default}
    elif default_factory is not None:
        defaults_kwargs = {"factory": default_factory}
    else:
        defaults_kwargs = {}

    return attrs.field(
        **defaults_kwargs,
        init=init,
        repr=repr,
        hash=hash,
        eq=compare,
        order=compare,
        metadata=metadata,
        kw_only=kw_only,
        converter=converter,  # type: ignore[arg-type]
        validator=validator,  # type: ignore[arg-type]
    )


coerced_field = functools.partial(field, converter="coerce")
"""Field definition function using ``converter="coerce"``."""


def validator(name: str) -> Callable[[FieldValidator], FieldValidator]:
    """Define a custom field validator for a specific field (decorator function).

    Arguments:
        name: Name of the field to be validated by the decorated function.

    The decorated functions should have the following signature:
    ``def _validator_function(self, attribute, value):``
    where ``self`` will be the model instance being validated, ``attribute``
    the definition information of the attribute (the value of
    ``__datamodel_fields__.field_name``) and ``value`` the actual value
    received for this field.
    """
    assert isinstance(name, str)

    def _field_validator_maker(func: FieldValidator) -> FieldValidator:
        names = getattr(func, _FIELD_VALIDATOR_TAG, ())
        setattr(func, _FIELD_VALIDATOR_TAG, tuple([*names, name]))
        return func

    return _field_validator_maker


_RV = TypeVar("_RV", bound=RootValidator)


def root_validator(cls_method: _RV, /) -> _RV:
    """Define a custom root validator (decorator function).

    The decorated functions should have the following signature:
    ``def _root_validator_function(cls, instance):``
    where ``cls`` will be the class of the model and ``instance`` the
    actual instance being validated.
    """
    setattr(cls_method, _ROOT_VALIDATOR_TAG, None)
    return cls_method


# -- Utils --
def is_datamodel(obj: Any) -> bool:
    """Return True if `obj` is a Data Model class or an instance of a Data Model."""
    cls = obj if isinstance(obj, type) else obj.__class__
    return hasattr(cls, MODEL_FIELD_DEFINITIONS_ATTR)


def is_generic_datamodel_class(cls: Type) -> bool:
    """Return ``True`` if `obj` is a generic Data Model class with type parameters."""
    assert isinstance(cls, type)
    return is_datamodel(cls) and xtyping.has_type_parameters(cls)


def get_fields(model: Union[DataModel, Type[DataModel]]) -> utils.FrozenNamespace:
    """Return the field meta-information of a Data Model.

    Arguments:
        model: A Data Model class or instance.

    Examples:
        >>> from typing import List
        >>> @datamodel
        ... class Model:
        ...     name: str
        ...     amount: int = 1
        ...     numbers: List[float] = field(default_factory=list)
        >>> fields(Model)  # doctest:+ELLIPSIS
        FrozenNamespace(...name=Attribute(name='name', default=NOTHING, ...

    """  # noqa: RST201  # doctest conventions confuse RST validator
    if not is_datamodel(model):
        raise TypeError(f"Invalid datamodel instance or class: '{model}'.")
    if not isinstance(model, type):
        model = model.__class__

    ns = getattr(model, MODEL_FIELD_DEFINITIONS_ATTR)
    assert isinstance(ns, utils.FrozenNamespace)
    return ns


fields = get_fields


def asdict(
    instance: DataModel,
    *,
    value_serializer: Optional[Callable[[Type[DataModel], Attribute, Any], Any]] = None,
) -> Dict[str, Any]:
    """Return the contents of a Data Model instance as a new mapping from field names to values.

    Arguments:
        instance: Data Model instance.

    Keyword Arguments:
        value_serializer: A hook that is called for every attribute or dict key/value and must
            return the (updated) value.

    Examples:
        >>> @datamodel
        ... class C:
        ...     x: int
        ...     y: int
        >>> c = C(x=1, y=2)
        >>> assert asdict(c) == {'x': 1, 'y': 2}
    """  # noqa: RST301  # sphinx.napoleon conventions confuse RST validator
    if not is_datamodel(instance) or isinstance(instance, type):
        raise TypeError(f"Invalid datamodel instance: '{instance}'.")
    return attrs.asdict(instance, value_serializer=value_serializer)


def astuple(instance: DataModel) -> Tuple[Any, ...]:
    """Return the contents of a Data Model instance as a new tuple of field values.

    Arguments:
        instance: Data Model instance.

    Examples:
        >>> @datamodel
        ... class C:
        ...     x: int
        ...     y: int
        >>> c = C(x=1, y=2)
        >>> assert astuple(c) == (1, 2)

    """
    if not is_datamodel(instance) or isinstance(instance, type):
        raise TypeError(f"Invalid datamodel instance: '{instance}'.")
    return attrs.astuple(instance, recurse=True)


evolve = attrs.evolve
"""Create a new instance, based on inst with changes applied."""

validate = attrs.validate
"""Validate all attributes on inst that have a validator."""

_DataModelT = TypeVar("_DataModelT", bound=DataModel)


def update_forward_refs(
    model_cls: Type[_DataModelT],
    localns: Optional[Dict[str, Any]] = None,
) -> Type[_DataModelT]:
    """Update Data Model class meta-information replacing forwarded type annotations with actual types.

    Arguments:
        localns: locals ``dict`` used in the evaluation of the annotations
            (globals are automatically taken from ``model.__module__``).

    Returns:
        The provided class (so it can be used as a decorator too).

    """
    if not (isinstance(model_cls, type) and is_datamodel(model_cls)):
        raise TypeError(f"Invalid datamodel class: '{model_cls}'.")

    # attrs.resolve_types() caches the exact class (in the MRO) whose types have been already resolved
    if getattr(model_cls, "__attrs_types_resolved__", None) != model_cls:
        fields = list(model_cls.__datamodel_fields__.keys())
        current_datamodel_fields = getattr(model_cls, MODEL_FIELD_DEFINITIONS_ATTR)

        for field_name in fields:
            if not hasattr(current_datamodel_fields, field_name):
                raise ValueError(f"{model_cls} does not contain a field named '{field_name}'.")

            field_attr = getattr(current_datamodel_fields, field_name)

            if isinstance(field_attr.type, ForwardRef):
                try:
                    actual_type = xtyping.eval_forward_ref(
                        field_attr.type,
                        sys.modules[model_cls.__module__].__dict__,
                        localns,
                        include_extras=True,
                    )
                    object.__setattr__(field_attr, "type", actual_type)
                except Exception as error:
                    raise TypeError(
                        f"Unexpected error trying to solve '{field_name}' field annotation ('{getattr(field_attr, 'type', None)}')"
                    ) from error

            # update resolved cache as attrs.resolve_types() would do
            model_cls.__attrs_types_resolved__ = model_cls  # type: ignore[attr-defined]  # adding a new class attribute

    return model_cls


def concretize(
    datamodel_cls: Type[GenericDataModelT],
    /,
    *type_args: Type,
    class_name: Optional[str] = None,
    module: Optional[str] = None,
    support_pickling: bool = True,  # noqa
    overwrite_definition: bool = True,
) -> Type[DataModelT]:
    """Generate a new concrete subclass of a generic Data Model.

    Arguments:
        datamodel_cls: Generic Data Model to be subclassed.
        type_args: Type definitions replacing the `TypeVars` in
            ``datamodel_cls.__parameters__``.

    Keyword Arguments:
        class_name: Name of the new concrete class. The default value is the
            same of the generic Data Model replacing the `TypeVars` by the provided
            `type_args` in the name.
        module: Value of the ``__module__`` attribute of the new class.
            The default value is the name of the module containing the generic Data Model.
        support_pickling: If ``True``, support for pickling will be added
            by actually inserting the new class into the target `module`.
        overwrite_definition: If ``True``, a previous definition of the class in
            the target module will be overwritten.

    """  # noqa: RST301  # doctest conventions confuse RST validator
    concrete_cls: Type[DataModelT] = _make_concrete_with_cache(
        datamodel_cls, *type_args, class_name=class_name, module=module
    )
    assert isinstance(concrete_cls, type) and is_datamodel(concrete_cls)

    # For pickling to work, the new class has to be added to the proper module
    if support_pickling:
        class_name = concrete_cls.__name__
        reference_module_globals = sys.modules[concrete_cls.__module__].__dict__
        if not (cls_in_module := (class_name in reference_module_globals)) or overwrite_definition:
            reference_module_globals[class_name] = concrete_cls
        elif cls_in_module and reference_module_globals[class_name] is not concrete_cls:
            warnings.warn(
                RuntimeWarning(
                    f"Existing '{class_name}' symbol in module '{module}' contains a reference"
                    "to a different object."
                )
            )

    return concrete_cls


# -- Helpers --
def _collect_field_validators(cls: Type) -> Dict[str, FieldValidator]:
    result = {}
    for member in cls.__dict__.values():
        if hasattr(member, _FIELD_VALIDATOR_TAG):
            for field_name in getattr(member, _FIELD_VALIDATOR_TAG):
                result[field_name] = member
            delattr(member, _FIELD_VALIDATOR_TAG)

    return result


def _collect_root_validators(cls: Type) -> List[RootValidator]:
    result = []
    for base in reversed(cls.__mro__[1:]):
        for validator in getattr(base, MODEL_ROOT_VALIDATORS_ATTR, []):
            if validator not in result:
                result.append(validator)

    for member in cls.__dict__.values():
        if hasattr(member, _ROOT_VALIDATOR_TAG):
            result.append(member)
            delattr(member, _ROOT_VALIDATOR_TAG)

    return result


def _get_attribute_from_bases(
    name: str, mro: Tuple[Type, ...], annotations: Optional[Dict[str, Any]] = None
) -> Optional[Attribute]:
    for base in mro:
        for base_field_attrib in getattr(base, "__attrs_attrs__", []):
            if base_field_attrib.name == name:
                if annotations is not None:
                    annotations[name] = base.__annotations__[name]
                return cast(Attribute, base_field_attrib)

    return None


def _substitute_typevars(
    type_hint: Type, type_params_map: Mapping[TypeVar, Union[Type, TypeVar]]
) -> Tuple[Union[Type, TypeVar], bool]:
    if isinstance(type_hint, typing.TypeVar):
        assert type_hint in type_params_map
        return type_params_map[type_hint], True
    elif getattr(type_hint, "__parameters__", []):
        return type_hint[tuple(type_params_map[tp] for tp in type_hint.__parameters__)], True
        # TODO(egparedes): WIP fix for partial specialization
        # # Type hint is a generic model: replace all the concretized type vars
        # noqa: e800 replaced = False
        # noqa: e800 new_args = []
        # noqa: e800 for tp in type_hint.__parameters__:
        # noqa: e800     if tp in type_params_map:
        # noqa: e800         new_args.append(type_params_map[tp])
        # noqa: e800         replaced = True
        # noqa: e800     else:
        # noqa: e800         new_args.append(type_params_map[tp])
        # noqa: e800 return type_hint[tuple(new_args)], replaced
    else:
        return type_hint, False


def _make_counting_attr_from_attribute(
    field_attrib: Attribute, *, include_type: bool = False, **kwargs: Any
) -> Any:  # attr.s lies a bit in some typing definitons
    args = [
        "default",
        "validator",
        "repr",
        "eq",
        "order",
        "hash",
        "init",
        "metadata",
        "converter",
        "kw_only",
        "on_setattr",
    ]
    if include_type:
        args.append("type")

    result = attr.ib(**{key: getattr(field_attrib, key) for key in args}, **kwargs)
    for key in ("eq_key", "order_key"):
        object.__setattr__(result, key, getattr(field_attrib, key))

    return result


def _make_post_init(has_post_init: bool) -> Callable[[DataModel], None]:
    # Duplicated code to facilitate the source inspection of the generated `__init__()` method
    if has_post_init:

        def __attrs_post_init__(self: DataModel) -> None:
            if attr._config._run_validators is True:  # type: ignore[attr-defined]  # attr._config is not visible for mypy
                for validator in self.__datamodel_root_validators__:
                    validator.__get__(self)(self)

            self.__post_init__()

    else:

        def __attrs_post_init__(self: DataModel) -> None:
            if attr._config._run_validators is True:  # type: ignore[attr-defined]  # attr._config is not visible for mypy
                for validator in type(self).__datamodel_root_validators__:
                    validator.__get__(self)(self)

    setattr(__attrs_post_init__, _DATAMODEL_TAG, True)

    return __attrs_post_init__


def _make_devtools_pretty() -> Callable[
    [DataModel, Callable[[Any], Any]], Generator[Any, None, None]
]:
    def __pretty__(
        self: DataModel, fmt: Callable[[Any], Any], **kwargs: Any
    ) -> Generator[Any, None, None]:
        """Provide a human readable representation for `devtools <https://python-devtools.helpmanual.io/>`_.

        Note:
            Adapted from `pydantic <https://github.com/samuelcolvin/pydantic>`_.
        """
        yield self.__class__.__name__ + "("
        yield 1
        for name in self.__datamodel_fields__.keys():
            yield name + "="
            yield fmt(getattr(self, name))
            yield ","
            yield 0
        yield -1
        yield ")"

    return __pretty__


def _make_data_model_class_getitem() -> classmethod:
    def __class_getitem__(
        cls: Type[GenericDataModelT], args: Union[Type, Tuple[Type]]
    ) -> Type[DataModelT] | Type[GenericDataModelT]:
        """Return an instance compatible with aliases created by :class:`typing.Generic` classes.

        See :class:`GenericDataModelAlias` for further information.
        """
        type_args: Tuple[Type] = args if isinstance(args, tuple) else (args,)
        concrete_cls: Type[DataModelT] = concretize(cls, *type_args)
        res = xtyping.StdGenericAliasType(concrete_cls, type_args)
        if sys.version_info < (3, 9):
            # in Python 3.8, xtyping.StdGenericAliasType (aka typing._GenericAlias)
            # does not copy all required `__dict__` entries, so do it manually
            for k, v in concrete_cls.__dict__.items():
                if k not in res.__dict__:
                    res.__dict__[k] = v
        return res

    return classmethod(__class_getitem__)


def _make_type_converter(type_annotation: TypeAnnotation, name: str) -> TypeConverter[_T]:
    # TODO(egparedes): if a "typing tree" structure is implemented, refactor this code as a tree traversal.
    #
    if xtyping.is_actual_type(type_annotation) and not isinstance(None, type_annotation):
        assert not xtyping.get_args(type_annotation)
        assert isinstance(type_annotation, type)

        def _type_converter(value: Any) -> _T:
            try:
                return (
                    value
                    if isinstance(value, type_annotation)  # type: ignore[arg-type]
                    else type_annotation(value)  # type: ignore
                )
            except Exception as error:
                raise TypeError(
                    f"Error during coertion of given value '{value}' for field '{name}'."
                ) from error

        return _type_converter

    if isinstance(type_annotation, TypeVar):
        return (
            _make_type_converter(type_annotation.__bound__, name)
            if type_annotation.__bound__
            else toolz.identity
        )

    if type_annotation is Any:
        return toolz.identity

    origin_type = xtyping.get_origin(type_annotation)

    if (
        origin_type is xtyping.Union
        and type(None) in (args := xtyping.get_args(type_annotation))
        and len(args) == 2
    ):
        # Optional type
        _inner_type_converter: TypeConverter[_T] = _make_type_converter(args[0], name)

        return cast(TypeConverter[_T], lambda x: x if x is None else _inner_type_converter(x))

    if xtyping.is_actual_type(origin_type):
        return _make_type_converter(origin_type, name)

    raise exceptions.EveTypeError(
        f"Automatic type coertion for {type_annotation} types is not supported."
    )


_KNOWN_MUTABLE_TYPES: Final = (list, dict, set)


def _make_datamodel(  # noqa: C901  # too complex but still readable and documented
    cls: Type[_T],
    *,
    repr: bool,  # noqa: A002   # shadowing 'repr' python builtin
    eq: bool,
    order: bool,
    unsafe_hash: bool,
    frozen: bool | Literal["strict"],
    match_args: bool,
    kw_only: bool,
    slots: bool,
    coerce: bool,
    generic: bool | Literal["True_no_checks"],
    type_validation_factory: Optional[FieldTypeValidatorFactory],
    _stacklevel_offset: int = 0,
) -> Type[_T]:
    """Actual implementation of the Data Model creation.

    See :func:`datamodel` for the description of the parameters.

    """
    mro_bases: Tuple[Type, ...] = cls.__mro__[1:]

    if "__annotations__" not in cls.__dict__:
        cls.__annotations__ = {}
    annotations = cls.__dict__["__annotations__"]
    resolved_annotations = xtyping.get_partial_type_hints(cls)
    annotations_with_extras = xtyping.get_partial_type_hints(cls, include_extras=True)

    frozen, strict_frozen = (True, True) if frozen == "strict" else (frozen, False)

    # Create attrib definitions with automatic type validators and converters
    # for the annotated fields. The keys in the original annotations are used for
    # iteration, since the resolved annotations also contain superclasses' annotations
    for key in annotations:
        type_hint = annotations[key] = resolved_annotations[key]

        # Skip members annotated as class variables
        if type_hint is ClassVar or xtyping.get_origin(type_hint) is ClassVar:
            continue

        if xtyping.get_origin(annotations_with_extras[key]) == xtyping.Annotated:
            _, *type_extras = xtyping.get_args(annotations_with_extras[key])
        else:
            type_extras = []

        coerce_field = coerce or _COERCED_TYPE_TAG in type_extras
        qualified_field_name = f"{cls.__name__}.{key}"

        # Create type validator if validation is enabled
        if type_validation_factory is None or _UNCHECKED_TYPE_TAG in type_extras:
            type_validator = lambda a, b, c: None  # noqa: E731
        else:
            type_validator = type_validation_factory(type_hint, qualified_field_name)

        # Declare an attrs.field() for this field with the right options.
        # There are different cases depending on the current value of the attribute in the class dict.
        attr_value_in_cls = cls.__dict__.get(key, NOTHING)
        if isinstance(attr_value_in_cls, attr._make._CountingAttr):  # type: ignore[attr-defined]  # _make is private
            # A field() function has already been used to customize the definition of the field.
            # In this case, we need to:
            #  - prepend the type validator to the list of provided validators (if any)
            #  - add the converter if the field needs to be converted and another custom converter
            #      has not been defined
            attr_value_in_cls._validator = (
                type_validator
                if attr_value_in_cls._validator is None
                else attr._make.and_(type_validator, attr_value_in_cls._validator)  # type: ignore[attr-defined]  # attr._make is not visible for mypy
            )
            coerce_field = coerce_field or attr_value_in_cls.converter == "coerce"
            if coerce_field:
                if attr_value_in_cls.converter in (None, "coerce"):
                    attr_value_in_cls.converter = _make_type_converter(
                        type_hint, qualified_field_name
                    )
                else:
                    raise TypeError(
                        f"Impossible to add automatic type converter to field '{qualified_field_name}' "
                        "which already defines a custom converter."
                    )

        else:
            # Create field converter if automatic coertion is enabled
            converter: TypeConverter = cast(
                TypeConverter,
                _make_type_converter(type_hint, qualified_field_name) if coerce_field else None,
            )
            if attr_value_in_cls is NOTHING:
                # The field has no definition in the class dict, it's only an annotation
                setattr(
                    cls,
                    key,
                    attrs.field(converter=converter, validator=type_validator),
                )

            else:
                # The field contains the default value in the class dict
                if isinstance(attr_value_in_cls, _KNOWN_MUTABLE_TYPES):
                    warnings.warn(
                        f"'{attr_value_in_cls.__class__.__name__}' value used as default in '{cls.__name__}.{key}'.\n"
                        "Mutable types should not defbe normally used as field defaults (use 'default_factory' instead).",
                        stacklevel=_stacklevel_offset + 2,
                    )
                setattr(
                    cls,
                    key,
                    attrs.field(
                        converter=converter, default=attr_value_in_cls, validator=type_validator
                    ),
                )

    # All fields should be annotated with type hints
    num_attrs = 0
    for key, value in cls.__dict__.items():
        if isinstance(value, attr._make._CountingAttr):  # type: ignore[attr-defined]  # attr._make is not visible for mypy
            num_attrs += 1
            if (
                key not in annotations
                and xtyping.get_origin(resolved_annotations.get(key, None)) is not ClassVar
            ):
                raise TypeError(f"Missing type annotation in '{key}' field.")

    # Validator processing
    root_validators = _collect_root_validators(cls)
    field_validators = _collect_field_validators(cls)

    for qualified_field_name, field_validator in field_validators.items():
        field_c_attr = cls.__dict__.get(qualified_field_name, None)
        if not field_c_attr:
            # Field has not been defined in the current class namespace,
            # look for field definition in the base classes.
            base_field_attr = _get_attribute_from_bases(
                qualified_field_name, mro_bases, annotations
            )
            if base_field_attr:
                # Create a new field in the current class cloning the existing
                # definition and add the new validator (attrs recommendation)
                field_c_attr = _make_counting_attr_from_attribute(
                    base_field_attr,
                )
                setattr(cls, qualified_field_name, field_c_attr)
            else:
                raise TypeError(
                    f"Validator assigned to non existing '{qualified_field_name}' field."
                )

        # Add field validator using field_attr.validator
        assert isinstance(field_c_attr, attr._make._CountingAttr)  # type: ignore[attr-defined]  # attr._make is not visible for mypy
        field_c_attr.validator(field_validator)

    setattr(cls, MODEL_ROOT_VALIDATORS_ATTR, tuple(root_validators))

    # Apply attrs.define() to enhance the class once all datamodels features
    # have been converted into attrs options
    if "__pre_init__" in cls.__dict__:
        if "__attrs_pre_init__" in cls.__dict__:
            raise TypeError(
                f"'{cls.__name__}' class contains custom '__attrs_pre_init__', which conflicts with `__pre_init__` ."
            )
        cls.__attrs_pre_init__ = cls.__pre_init__  # type: ignore[attr-defined]  # adding new attribute

    if "__attrs_post_init__" in cls.__dict__ and not hasattr(
        cls.__attrs_post_init__, _DATAMODEL_TAG  # type: ignore[attr-defined]  # mypy doesn't know about __attr_post_init__
    ):
        raise TypeError(f"'{cls.__name__}' class contains forbidden custom '__attrs_post_init__'.")
    cls.__attrs_post_init__ = _make_post_init(has_post_init="__post_init__" in cls.__dict__)  # type: ignore[attr-defined]  # adding new attribute

    if generic:
        if generic == "True_no_checks":
            # For the root GenericDataModel class, only set the attribute in __datamodel_params__ to True
            generic = True
        else:
            # For any other subclass, add the proper __class_getitem__ method
            if not issubclass(cls, (typing.Generic, xtyping.Generic)):  # type: ignore[arg-type]  # Generic is not considered a type
                raise TypeError(
                    f"'{cls.__name__}' cannot be converted to a GenericDataModel because it is not a generic class."
                )
            cls.__class_getitem__ = _make_data_model_class_getitem()  # type: ignore[attr-defined]  # adding new attribute

    # TODO(egparedes): consider the use of the field_transformer hook available in attrs:
    #   https://www.attrs.org/en/stable/extending.html#automatic-field-transformation-and-modification
    new_cls = attrs.define(  # type: ignore[call-overload]
        cls,
        auto_attribs=True,
        auto_detect=True,
        init=None,
        repr=None if repr else False,
        eq=None if eq else False,
        order=order,
        hash=None if not unsafe_hash else True,
        frozen=frozen,
        match_args=match_args,
        kw_only=kw_only,
        slots=slots,
    )
    assert (new_cls is cls) or slots

    # Final checks and postprocessing
    if strict_frozen:
        unhashable_fields = set()
        for f_attr in new_cls.__attrs_attrs__:
            if is_datamodel(f_attr.type):
                if getattr(f_attr.type, MODEL_PARAM_DEFINITIONS_ATTR).strict_frozen is True:
                    continue
            elif xtyping.is_hashable_type(f_attr.type):
                continue
            unhashable_fields.add(f_attr.name)

        if unhashable_fields:
            raise exceptions.EveTypeError(
                f"Some fields ({unhashable_fields}) can not be considered strictly immutable."
            )

    if "__attrs_init__" in new_cls.__dict__:
        assert "__auto_init__" not in cls.__dict__
        new_cls.__auto_init__ = new_cls.__attrs_init__

    new_cls.__pretty__ = _make_devtools_pretty()
    setattr(
        new_cls,
        MODEL_PARAM_DEFINITIONS_ATTR,
        utils.FrozenNamespace(
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
            strict_frozen=strict_frozen,
            match_args=match_args,
            kw_only=kw_only,
            slots=slots,
            coerce=coerce,
            generic=generic,
            type_validation_factory=type_validation_factory,
        ),
    )
    setattr(
        new_cls,
        MODEL_FIELD_DEFINITIONS_ATTR,
        utils.FrozenNamespace(**{f_attr.name: f_attr for f_attr in new_cls.__attrs_attrs__}),
    )
    new_cls.update_forward_refs = classmethod(update_forward_refs)

    return new_cls


@utils.optional_lru_cache(maxsize=None, typed=True)
def _make_concrete_with_cache(
    datamodel_cls: Type[GenericDataModelT],
    *type_args: Type,
    class_name: Optional[str] = None,
    module: Optional[str] = None,
) -> Type[DataModelT]:
    if not is_generic_datamodel_class(datamodel_cls):
        raise TypeError(f"'{datamodel_cls.__name__}' is not a generic model class.")
    for t in type_args:
        if not (
            isinstance(t, (type, type(None), xtyping.StdGenericAliasType))
            or (getattr(type(t), "__module__", None) in ("typing", "typing_extensions"))
        ):
            raise TypeError(
                f"Only 'type' and 'typing' definitions can be passed as arguments "
                f"to instantiate a generic model class (received: {type_args})."
            )
    if len(type_args) != len(datamodel_cls.__parameters__):
        raise TypeError(
            f"Instantiating '{datamodel_cls.__name__}' generic model with a wrong number of parameters "
            f"({len(type_args)} used, {len(datamodel_cls.__parameters__)} expected)."
        )

    # Replace field definitions with the new actual types for generic fields
    type_params_map = dict(zip(datamodel_cls.__parameters__, type_args))
    model_fields = getattr(datamodel_cls, MODEL_FIELD_DEFINITIONS_ATTR)
    new_annotations = {
        # TODO(egparedes): ?
        # noqa: e800 "__args__": "ClassVar[Tuple[Union[Type, TypeVar], ...]]",
        # noqa: e800 "__parameters__": "ClassVar[Tuple[TypeVar, ...]]",
    }

    new_field_c_attrs = {}
    for field_name, field_type in xtyping.get_type_hints(datamodel_cls).items():
        new_annotation, replaced = _substitute_typevars(field_type, type_params_map)
        if replaced:
            new_annotations[field_name] = new_annotation
            field_attrib = getattr(model_fields, field_name)
            new_field_c_attrs[field_name] = _make_counting_attr_from_attribute(field_attrib)

    # Create new concrete class
    if not class_name:
        arg_names = []
        for tp_var in datamodel_cls.__parameters__:
            arg_string = tp_var.__name__
            if tp_var in type_params_map:
                concrete_arg = type_params_map[tp_var]
                if isinstance(concrete_arg, type):
                    arg_string = concrete_arg.__name__
                else:
                    arg_string = utils.slugify(
                        str(concrete_arg).replace("typing.", "").replace("...", "ellipsis")
                    )

            arg_names.append(arg_string)

        class_name = f"{datamodel_cls.__name__}__{'_'.join(arg_names)}"

    namespace = {
        "__annotations__": new_annotations,
        "__module__": module if module else datamodel_cls.__module__,
        **new_field_c_attrs,
    }

    concrete_cls = type(class_name, (datamodel_cls,), namespace)
    assert concrete_cls.__module__ == module or not module

    if MODEL_FIELD_DEFINITIONS_ATTR not in concrete_cls.__dict__:
        # If original model does not inherit from GenericDataModel,
        # _make_datamodel() hasn't been called automatically, so call it now
        params = getattr(datamodel_cls, MODEL_PARAM_DEFINITIONS_ATTR)
        concrete_cls = _make_datamodel(
            concrete_cls,
            **{
                name: getattr(params, name)
                for name in ("repr", "eq", "order", "unsafe_hash", "frozen")
            },
        )

    return concrete_cls


if xtyping.TYPE_CHECKING:
    FrozenModel: TypeAlias = DataModel

else:

    class FrozenModel(DataModel, frozen=True):
        __slots__ = ()


if xtyping.TYPE_CHECKING:

    class GenericDataModel(GenericDataModelTP):
        @classmethod
        def __class_getitem__(
            cls: Type[GenericDataModelTP], args: Union[Type, Tuple[Type, ...]]
        ) -> Union[DataModelTP, GenericDataModelTP]:
            ...

else:

    class GenericDataModel(DataModel, __dm_opts=[_GENERIC_DATAMODEL_ROOT_DM_OPT]):
        __slots__ = ()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Data Model classes and related utils and validators.

Data Models can be considered as enhanced  `attrs <https://www.attrs.org>`_
/ `dataclasses <https://docs.python.org/3/library/dataclasses.html>`_ providing
additional features like automatic run-time type validation. Values assigned to fields
at initialization can be validated with automatic type checkings using the
field type definition. Custom field validation methods can also be added with
the :func:`validator` decorator, and global instance validation methods with
:func:`root_validator`.

The datamodels API tries to follow ``dataclasses`` API conventions when possible,
but it is not an exact copy. Implementation-wise, Data Model classes are just
customized ``attrs`` classes, and therefore external tools compatible with ``attrs``
classes should also work with Data Models classes.

A valid ``__init__`` method for the Data Model class is always generated. If the class
already defines a custom ``__init__`` method, the generated method will be named
``__auto_init__`` and should be called from the custom ``__init__`` to profit from
datamodels features. Additionally, if custom ``__pre_init__(self) -> None`` or
``__post_init__(self) -> None`` methods exist in the class, they will be automatically
called from the generated ``__init__`` before and after the instance creation.

The order of execution of the different components at initialization is the following:

    1. ``__init__()``.

        a.  If a custom ``__init__`` already exists in the class, it will not be overwritten.
            It is your responsability to call ``__auto_init__`` from there to obtain
            the described behavior.
        b.  If there is not custom ``__init__``, the one generated by datamodels
            will be called first.

    2. ``__pre_init__()``, if it exists.
    3. For each field, in the order it was defined:

        a. default factory (if any).
        b. converter (if any).

    4. all field validators.
    5. root validators following the reversed class MRO order.
    6. ``__post_init__()``, if it exists.


Examples:
    >>> @datamodel
    ... class SampleModel:
    ...     name: str
    ...     amount: int
    ...
    ...     @validator('name')
    ...     def _name_validator(self, attribute, value):
    ...         if len(value) < 3:
    ...             raise ValueError(
    ...                 f"Provided value '{value}' for '{attribute.name}' field is too short."
    ...             )

    >>> SampleModel("Some Name", 10)
    SampleModel(name='Some Name', amount=10)

    >>> SampleModel("A", 10)
    Traceback (most recent call last):
        ...
    ValueError: Provided value 'A' for 'name' field is too short.

    >>> class AnotherSampleModel(DataModel):
    ...     name: str
    ...     friends: List[str]
    ...
    ...     @root_validator
    ...     def _root_validator(cls, instance):
    ...         if instance.name in instance.friends:
    ...             raise ValueError("'name' value cannot appear in 'friends' list.")

    >>> AnotherSampleModel("Eve", ["Liam", "John"])
    AnotherSampleModel(name='Eve', friends=['Liam', 'John'])

    >>> AnotherSampleModel("Eve", ["Eve", "John"])
    Traceback (most recent call last):
        ...
    ValueError: 'name' value cannot appear in 'friends' list.

    >>> @datamodel
    ... class CustomModel:
    ...     value: float
    ...     num_instances: ClassVar[int] = 0
    ...
    ...     def __init__(self, a: int, b: int) -> None:
    ...         self.__auto_init__(a/b)
    ...
    ...     def __pre_init__(self) -> None:
    ...         self.__class__.num_instances += 1
    ...
    ...     def __post_init__(self) -> None:
    ...         print(f"Instance {self.num_instances} == {self.value}")

    >>> CustomModel(3, 2)
    Instance 1 == 1.5
    CustomModel(value=1.5)

"""

from . import core as core  # noqa: F401  # imported but unused
from . import validators as validators  # noqa: F401  # imported but unused
from .core import *  # noqa:   # star unused import



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Generic interface and implementations of run-time type validation for arbitrary values."""


from __future__ import annotations

import abc
import collections.abc
import dataclasses
import functools
import typing as _typing

from . import exceptions
from . import extended_typing as xtyping
from . import utils
from .extended_typing import (
    Any,
    Dict,
    Final,
    ForwardRef,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Type,
    TypeAnnotation,
    Union,
    cast,
    overload,
    runtime_checkable,
)


# Protocols
@runtime_checkable
class TypeValidator(Protocol):
    @abc.abstractmethod
    def __call__(
        self,
        value: Any,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        required: bool = True,
        **kwargs: Any,
    ) -> None:
        """Protocol for callables checking that ``value`` matches ``type_annotation``.

        Arguments:
            value: value to be checked against the typing annotation.
            type_annotation: a valid typing annotation.
            name: the name of the checked value (used for error messages).

        Keyword Arguments:
            globalns: globals dict used in the evaluation of the annotations.
            localns: locals dict used in the evaluation of the annotations.
            required: if ``True``, raise ``ValueError`` when provided type annotation is not supported.
            **kwargs: arbitrary implementation-defined arguments (e.g. for memoization).

        Raises:
            TypeError: if there is a type mismatch.
            ValueError: if ``required is True`` and ``type_annotation`` is not supported.
        """
        ...


class FixedTypeValidator(Protocol):
    @abc.abstractmethod
    def __call__(
        self,
        value: Any,
        **kwargs: Any,
    ) -> None:
        """Protocol for callables checking that ``value`` matches a fixed type_annotation.

        Arguments:
            value: value to be checked against the typing annotation.

        Keyword Arguments:
            **kwargs: arbitrary implementation-defined arguments (e.g. for memoization).

        Raises:
            TypeError: if there is a type mismatch.
        """
        ...


@runtime_checkable
class TypeValidatorFactory(Protocol):
    @overload
    def __call__(
        self,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        required: Literal[True] = True,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> FixedTypeValidator:
        ...

    @overload
    def __call__(  # noqa: F811  # redefinion of unused member
        self,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        required: bool = True,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Optional[FixedTypeValidator]:
        ...

    @abc.abstractmethod
    def __call__(  # noqa: F811  # redefinion of unused member
        self,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        required: bool = True,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Optional[FixedTypeValidator]:
        """Protocol for :class:`FixedTypeValidator`s.

        The arguments match the specification in :class:`TypeValidator`.

        Raises:
            TypeError: if there is a type mismatch.
            ValueError: if ``required is True`` and ``type_annotation`` is not supported.
        """
        ...


# Implementations
@dataclasses.dataclass(frozen=True)
class SimpleTypeValidatorFactory(TypeValidatorFactory):
    """A simple :class:`TypeValidatorFactory` implementation.

    Check :class:`FixedTypeValidator` and :class:`TypeValidatorFactory` for details.

    Keyword Arguments:
        strict_int (bool): do not accept ``bool`` values as ``int`` (default: ``True``).
    """

    @overload
    def __call__(
        self,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        required: Literal[True] = True,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> FixedTypeValidator:
        ...

    @overload
    def __call__(  # noqa: F811  # redefinion of unused member
        self,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        required: bool = True,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Optional[FixedTypeValidator]:
        ...

    def __call__(  # noqa: F811,C901  # redefinion of unused member / complex but well organized in cases
        self,
        type_annotation: TypeAnnotation,
        name: Optional[str] = None,
        *,
        required: bool = True,
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Optional[FixedTypeValidator]:

        # TODO(egparedes): if a "typing tree" structure is implemented, refactor this code as a tree traversal.
        #
        if name is None:
            name = "<value>"

        make_recursive = functools.partial(
            self.__call__, name=name, globalns=globalns, localns=localns, **kwargs
        )

        try:
            if type_annotation is None:
                type_annotation = type(None)

            # Non-generic types
            if xtyping.is_actual_type(type_annotation):
                assert not xtyping.get_args(type_annotation)
                if type_annotation is int and kwargs.get("strict_int", True):
                    return self.make_is_instance_of_int(name)
                else:
                    return self.make_is_instance_of(name, type_annotation)

            if isinstance(type_annotation, _typing.TypeVar):
                if type_annotation.__bound__:
                    return self.make_is_instance_of(name, type_annotation.__bound__)
                else:
                    return self._make_is_any(name)

            if isinstance(type_annotation, ForwardRef):
                return make_recursive(
                    xtyping.eval_forward_ref(type_annotation, globalns=globalns, localns=localns)
                )

            if xtyping.is_Any(type_annotation):
                return self._make_is_any(name)

            # Generic and parametrized type hints
            origin_type = xtyping.get_origin(type_annotation)
            type_args = xtyping.get_args(type_annotation)

            if origin_type is Literal:
                if len(type_args) == 1:
                    return self.make_is_literal(name, type_args[0])
                else:
                    return self.combine_validators_as_or(
                        name,
                        *(self.make_is_literal(name, a) for a in type_args),
                        error_type=TypeError,
                    )

            if origin_type is Union:
                has_none = False
                validators = []
                for t in type_args:
                    if t in (type(None), None):
                        has_none = True
                    else:
                        if (v := make_recursive(t)) is None:
                            raise exceptions.EveValueError(f"{t} type annotation is not supported.")
                        validators.append(v)

                validator = (
                    self.combine_validators_as_or(name, *validators)
                    if len(validators) > 1
                    else validators[0]
                )
                return self.combine_optional(name, validator) if has_none else validator

            if isinstance(origin_type, type):
                # Deal with generic collections
                if issubclass(origin_type, tuple):
                    if len(type_args) == 2 and (type_args[1] is Ellipsis):
                        # Tuple as an immutable sequence type (e.g. Tuple[int, ...])
                        if (member_validator := make_recursive(type_args[0])) is None:
                            raise exceptions.EveValueError(
                                f"{type_args[0]} type annotation is not supported."
                            )

                        return self.make_is_iterable_of(
                            name,
                            member_validator,
                            iterable_validator=self.make_is_instance_of(name, origin_type),
                        )

                    else:
                        # Tuple as a heterogeneous container (e.g. Tuple[int, float])
                        item_validators = []
                        for t in type_args:
                            if (v := make_recursive(t)) is None:
                                raise exceptions.EveValueError(
                                    f"{t} type annotation is not supported."
                                )
                            item_validators.append(v)

                        return self.make_is_tuple_of(name, tuple(item_validators), origin_type)

                if issubclass(origin_type, (collections.abc.Sequence, collections.abc.Set)):
                    assert len(type_args) == 1
                    if (member_validator := make_recursive(type_args[0])) is None:
                        raise exceptions.EveValueError(
                            f"{type_args[0]} type annotation is not supported."
                        )

                    return self.make_is_iterable_of(
                        name,
                        member_validator,
                        iterable_validator=self.make_is_instance_of(name, origin_type),
                    )

                if issubclass(origin_type, collections.abc.Mapping):
                    assert len(type_args) == 2
                    if (key_validator := make_recursive(type_args[0])) is None:
                        raise exceptions.EveValueError(
                            f"{type_args[0]} type annotation is not supported."
                        )
                    if (value_validator := make_recursive(type_args[1])) is None:
                        raise exceptions.EveValueError(
                            f"{type_args[1]} type annotation is not supported."
                        )

                    return self.make_is_mapping_of(
                        name,
                        key_validator,
                        value_validator,
                        mapping_validator=self.make_is_instance_of(name, origin_type),
                    )

                # Custom generic type: any regular (not datamodel) user-defined generic types like:
                #   class Foo(Generic[T]):
                #          ...
                #
                # Since this can be an arbitrary type (not something regular like a collection) there is
                # no way to check if the type parameter is verifed in the actual instance.
                # The only check can be done at run-time is to verify that the value is an instance of
                # the original type, completely ignoring the annotation. Ideally, the static type checker
                # can do a better job to try figure out if the type parameter is ok ...

                return make_recursive(origin_type)

            # TODO(egparedes): add support for signature checking in Callables
            raise exceptions.EveValueError(f"{type_annotation} type annotation is not supported.")

        except exceptions.EveValueError as error:
            if required:
                raise error

        assert bool(required) is False

        return None

    @staticmethod
    def _make_is_any(name: str) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for any type."""

        def _is_any(value: Any, **kwargs: Any) -> None:
            pass

        return _is_any

    @staticmethod
    def make_is_instance_of(name: str, type_: type) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for a specific type."""

        def _is_instance_of(value: Any, **kwargs: Any) -> None:
            if not isinstance(value, type_):
                raise TypeError(
                    f"'{name}' must be {type_} (got '{value}' which is a {type(value)})."
                )

        return _is_instance_of

    @staticmethod
    def make_is_instance_of_int(name: str) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for ``int`` values which fails with ``bool`` values."""

        def _is_instance_of_int(value: Any, **kwargs: Any) -> None:
            if not isinstance(value, int) or isinstance(value, bool):
                raise TypeError(f"'{name}' must be {int} (got '{value}' which is a {type(value)}).")

        return _is_instance_of_int

    @staticmethod
    def make_is_literal(name: str, literal_value: Any) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for a literal value."""
        if isinstance(literal_value, bool):

            def _is_literal(value: Any, **kwargs: Any) -> None:
                if value is not literal_value:
                    raise TypeError(
                        f"Provided value '{value}' for '{name}' does not match literal {literal_value}."
                    )

        else:

            def _is_literal(value: Any, **kwargs: Any) -> None:
                if value != literal_value:
                    raise TypeError(
                        f"Provided value '{value}' for '{name}' does not match literal {literal_value}."
                    )

        return _is_literal

    @staticmethod
    def make_is_tuple_of(
        name: str, item_validators: Sequence[FixedTypeValidator], tuple_type: type[tuple]
    ) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for tuple types."""

        def _is_tuple_of(value: Any, **kwargs: Any) -> None:
            if not isinstance(value, tuple_type):
                raise TypeError(
                    f"In '{name}' validation, got '{value}' which is a {type(value)} instead of {tuple_type}."
                )
            if len(value) != len(item_validators):
                raise TypeError(
                    f"In '{name}' validation, got '{value}' tuple which contains {len(value)} elements instead of {len(item_validators)}."
                )

            _i = None
            item_value = ""
            try:
                for _i, (item_value, item_validator) in enumerate(zip(value, item_validators)):
                    item_validator(item_value)
            except Exception as e:
                raise TypeError(
                    f"In '{name}' validation, tuple '{value}' contains invalid value '{item_value}' at position {_i}."
                ) from e

        return _is_tuple_of

    @staticmethod
    def make_is_iterable_of(
        name: str,
        member_validator: FixedTypeValidator,
        iterable_validator: Optional[FixedTypeValidator] = None,
    ) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for deep checks of typed iterables."""

        def _is_iterable_of(value: Any, **kwargs: Any) -> None:
            if iterable_validator is not None:
                iterable_validator(value, **kwargs)

            for member in value:
                member_validator(member, **kwargs)

        return _is_iterable_of

    @staticmethod
    def make_is_mapping_of(
        name: str,
        key_validator: FixedTypeValidator,
        value_validator: FixedTypeValidator,
        mapping_validator: Optional[FixedTypeValidator] = None,
    ) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for deep checks of typed mappings."""

        def _is_mapping_of(value: Any, **kwargs: Any) -> None:
            if mapping_validator is not None:
                mapping_validator(value, **kwargs)

            for k in value:
                key_validator(k, **kwargs)
                value_validator(value[k], **kwargs)

        return _is_mapping_of

    @staticmethod
    def combine_optional(name: str, actual_validator: FixedTypeValidator) -> FixedTypeValidator:
        """Create an ``FixedTypeValidator`` validator for an optional constraint."""

        def _is_optional(value: Any, **kwargs: Any) -> None:
            if value is not None:
                actual_validator(value, **kwargs)

        return _is_optional

    @staticmethod
    def combine_validators_as_or(
        name: str, *validators: FixedTypeValidator, error_type: Type[Exception] = TypeError
    ) -> FixedTypeValidator:
        def _combined_validator(value: Any, **kwargs: Any) -> None:
            for v in validators:
                try:
                    v(value, **kwargs)
                    break
                except Exception:
                    pass
            else:
                raise error_type(
                    f"In '{name}' validation, provided value '{value}' fails for all the possible validators."
                )

        return _combined_validator


simple_type_validator_factory: Final = cast(
    TypeValidatorFactory,
    utils.optional_lru_cache(SimpleTypeValidatorFactory(), typed=True),
)
"""Public (with optional cache) entry point for :class:`SimpleTypeValidatorFactory`."""


def simple_type_validator(
    value: Any,
    type_annotation: TypeAnnotation,
    name: Optional[str] = None,
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    required: bool = True,
    **kwargs: Any,
) -> None:
    """Check that value satisfies a type definition (a simple :class:`TypeValidator` implementation).

    Check :class:`TypeValidator` and :class:`SimpleTypeValidatorFactory` for details.

    Keyword Arguments:
        strict_int (bool): do not accept ``bool`` values as ``int`` (default: ``True``).
    """
    type_validator: Optional[FixedTypeValidator] = simple_type_validator_factory(
        type_annotation, name=name, globalns=globalns, localns=localns, required=required, **kwargs
    )
    if type_validator is not None:
        type_validator(value, **kwargs)


# TODO(egparedes): add other implementations for advanced 3rd-party validators
#   e.g. 'typeguard' and specially 'beartype'



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Global configuration of test generation and execution with pytest."""


import pytest

from . import definitions


NODE_MAKERS = []
FROZEN_NODE_MAKERS = []
INVALID_NODE_MAKERS = []


# Automatic creation of pytest fixtures from maker functions in .definitions
for key, value in definitions.__dict__.items():
    if key.startswith("make_"):
        name = key[5:]
        exec(
            f"""
@pytest.fixture
def {name}_maker():
    yield definitions.make_{name}

@pytest.fixture
def {name}({name}_maker):
    yield {name}_maker()

@pytest.fixture
def fixed_{name}({name}_maker):
    yield {name}_maker(fixed=True)
"""
        )

        if "_node" in key:
            if "_invalid" in key:
                INVALID_NODE_MAKERS.append(value)
            elif "_frozen" in key:
                FROZEN_NODE_MAKERS.append(value)
            else:
                NODE_MAKERS.append(value)


@pytest.fixture(params=NODE_MAKERS)
def sample_node_maker(request):
    return request.param


@pytest.fixture(params=NODE_MAKERS)
def sample_node(request):
    return request.param()


@pytest.fixture(params=FROZEN_NODE_MAKERS)
def frozen_sample_node_maker(request):
    return request.param


@pytest.fixture(params=FROZEN_NODE_MAKERS)
def frozen_sample_node(request):
    return request.param()


@pytest.fixture(params=INVALID_NODE_MAKERS)
def invalid_sample_node_maker(request):
    return request.param



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses
import re
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generic, List, Optional, Set, Tuple, TypeVar, cast

import gt4py.eve.utils
from gt4py.cartesian.gtc import common, oir
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.cartesian.gtc.passes.horizontal_masks import mask_overlap_with_extent


OffsetT = TypeVar("OffsetT")

GeneralOffsetTuple = Tuple[int, int, Optional[int]]

_digits_at_end_pattern = re.compile(r"[0-9]+$")
_generated_name_pattern = re.compile(r".+_gen_[0-9]+")


@dataclass(frozen=True)
class GenericAccess(Generic[OffsetT]):
    field: str
    offset: OffsetT
    is_write: bool
    data_index: List[oir.Expr] = dataclasses.field(default_factory=list)
    horizontal_mask: Optional[common.HorizontalMask] = None

    @property
    def is_read(self) -> bool:
        return not self.is_write

    def to_extent(self, horizontal_extent: Extent) -> Optional[Extent]:
        """
        Convert the access to an extent provided a horizontal extent for the access.

        This returns None if no overlap exists between the horizontal mask and interval.
        """
        offset_as_extent = Extent.from_offset(cast(Tuple[int, int, int], self.offset)[:2])
        zeros = Extent.zeros(ndims=2)
        if self.horizontal_mask:
            if dist_from_edge := mask_overlap_with_extent(self.horizontal_mask, horizontal_extent):
                return ((horizontal_extent - dist_from_edge) + offset_as_extent) | zeros
            else:
                return None
        else:
            return horizontal_extent + offset_as_extent


class CartesianAccess(GenericAccess[Tuple[int, int, int]]):
    pass


class GeneralAccess(GenericAccess[GeneralOffsetTuple]):
    pass


AccessT = TypeVar("AccessT", bound=GenericAccess)


class AccessCollector(gt4py.eve.NodeVisitor):
    """Collects all field accesses and corresponding offsets."""

    def visit_FieldAccess(
        self,
        node: oir.FieldAccess,
        *,
        accesses: List[GeneralAccess],
        is_write: bool,
        horizontal_mask: Optional[common.HorizontalMask] = None,
        **kwargs: Any,
    ) -> None:
        self.generic_visit(node, accesses=accesses, is_write=is_write, **kwargs)
        offsets = node.offset.to_dict()
        accesses.append(
            GeneralAccess(
                field=node.name,
                offset=(cast(int, offsets["i"]), cast(int, offsets["j"]), offsets["k"]),
                data_index=node.data_index,
                is_write=is_write,
                horizontal_mask=horizontal_mask,
            )
        )

    def visit_AssignStmt(
        self,
        node: oir.AssignStmt,
        **kwargs: Any,
    ) -> None:
        self.visit(node.right, is_write=False, **kwargs)
        self.visit(node.left, is_write=True, **kwargs)

    def visit_MaskStmt(self, node: oir.MaskStmt, **kwargs: Any) -> None:

        self.visit(node.mask, is_write=False, **kwargs)
        self.visit(node.body, **kwargs)

    def visit_While(self, node: oir.While, **kwargs: Any) -> None:
        self.visit(node.cond, is_write=False, **kwargs)
        self.visit(node.body, **kwargs)

    def visit_HorizontalRestriction(self, node: oir.HorizontalRestriction, **kwargs: Any) -> None:
        self.visit(node.body, horizontal_mask=node.mask, **kwargs)

    @dataclass
    class GenericAccessCollection(Generic[AccessT, OffsetT]):
        _ordered_accesses: List[AccessT]

        @staticmethod
        def _offset_dict(accesses: gt4py.eve.utils.XIterable) -> Dict[str, Set[OffsetT]]:
            return accesses.reduceby(
                lambda acc, x: acc | {x.offset}, "field", init=set(), as_dict=True
            )

        def offsets(self) -> Dict[str, Set[OffsetT]]:
            """Get a dictionary, mapping all accessed fields' names to sets of offset tuples."""
            return self._offset_dict(gt4py.eve.utils.XIterable(self._ordered_accesses))

        def read_offsets(self) -> Dict[str, Set[OffsetT]]:
            """Get a dictionary, mapping read fields' names to sets of offset tuples."""
            return self._offset_dict(
                gt4py.eve.utils.XIterable(self._ordered_accesses).filter(lambda x: x.is_read)
            )

        def read_accesses(self) -> List[AccessT]:
            """Get the sub-list of read accesses."""
            return list(
                gt4py.eve.utils.XIterable(self._ordered_accesses).filter(lambda x: x.is_read)
            )

        def write_offsets(self) -> Dict[str, Set[OffsetT]]:
            """Get a dictionary, mapping written fields' names to sets of offset tuples."""
            return self._offset_dict(
                gt4py.eve.utils.XIterable(self._ordered_accesses).filter(lambda x: x.is_write)
            )

        def write_accesses(self) -> List[AccessT]:
            """Get the sub-list of write accesses."""
            return list(
                gt4py.eve.utils.XIterable(self._ordered_accesses).filter(lambda x: x.is_write)
            )

        def fields(self) -> Set[str]:
            """Get a set of all accessed fields' names."""
            return {acc.field for acc in self._ordered_accesses}

        def read_fields(self) -> Set[str]:
            """Get a set of all read fields' names."""
            return {acc.field for acc in self._ordered_accesses if acc.is_read}

        def write_fields(self) -> Set[str]:
            """Get a set of all written fields' names."""
            return {acc.field for acc in self._ordered_accesses if acc.is_write}

        def ordered_accesses(self) -> List[AccessT]:
            """Get a list of ordered accesses."""
            return self._ordered_accesses

    class CartesianAccessCollection(GenericAccessCollection[CartesianAccess, Tuple[int, int, int]]):
        pass

    class GeneralAccessCollection(GenericAccessCollection[GeneralAccess, GeneralOffsetTuple]):
        def cartesian_accesses(self) -> "AccessCollector.CartesianAccessCollection":
            return AccessCollector.CartesianAccessCollection(
                [
                    CartesianAccess(
                        field=acc.field,
                        offset=cast(Tuple[int, int, int], acc.offset),
                        data_index=acc.data_index,
                        is_write=acc.is_write,
                    )
                    for acc in self._ordered_accesses
                    if acc.offset[2] is not None
                ]
            )

        def has_variable_access(self) -> bool:
            return any(acc.offset[2] is None for acc in self._ordered_accesses)

    @classmethod
    def apply(
        cls, node: gt4py.eve.RootNode, **kwargs: Any
    ) -> "AccessCollector.GeneralAccessCollection":
        result = cls.GeneralAccessCollection([])
        cls().visit(node, accesses=result._ordered_accesses, **kwargs)
        return result


def symbol_name_creator(used_names: Set[str]) -> Callable[[str], str]:
    """Create a function that generates symbol names that are not already in use.

    Args:
        used_names: Symbol names that are already in use and thus should not be generated.
                    NOTE: `used_names` will be modified to contain all generated symbols.

    Returns:
        A callable to generate new unique symbol names.
    """

    def increment_string_suffix(s: str) -> str:
        if not _generated_name_pattern.match(s):
            return s + "_gen_0"
        return _digits_at_end_pattern.sub(lambda n: str(int(n.group()) + 1), s)

    def new_symbol_name(name: str) -> str:
        while name in used_names:
            name = increment_string_suffix(name)
        used_names.add(name)
        return name

    return new_symbol_name


def collect_symbol_names(node: gt4py.eve.RootNode) -> Set[str]:
    return (
        gt4py.eve.walk_values(node)
        .if_isinstance(gt4py.eve.SymbolTableTrait)
        .getattr("annex")
        .getattr("symtable")
        .reduce(lambda names, symtable: names.union(symtable.keys()), init=set())
    )


class StencilExtentComputer(gt4py.eve.NodeVisitor):
    @dataclass
    class Context:
        fields: Dict[str, Extent] = dataclasses.field(default_factory=dict)
        blocks: Dict[int, Extent] = dataclasses.field(default_factory=dict)

    def __init__(self, add_k: bool = False):
        self.add_k = add_k
        self.zero_extent = Extent.zeros(ndims=2)

    def visit_Stencil(self, node: oir.Stencil) -> "Context":
        ctx = self.Context()
        for vloop in reversed(node.vertical_loops):
            self.visit(vloop, ctx=ctx)

        if self.add_k:
            ctx.fields = {name: Extent(*extent, (0, 0)) for name, extent in ctx.fields.items()}

        for name in (p.name for p in node.params if p.name not in ctx.fields):
            ctx.fields[name] = self.zero_extent

        return ctx

    def visit_VerticalLoopSection(self, node: oir.VerticalLoopSection, **kwargs: Any) -> None:
        for hexec in reversed(node.horizontal_executions):
            self.visit(hexec, **kwargs)

    def visit_HorizontalExecution(self, node: oir.HorizontalExecution, *, ctx: Context) -> None:
        results = AccessCollector.apply(node)

        horizontal_extent = self.zero_extent
        for access in (acc for acc in results.ordered_accesses() if acc.is_write):
            horizontal_extent |= ctx.fields.setdefault(access.field, self.zero_extent)
        ctx.blocks[id(node)] = horizontal_extent

        for access in results.ordered_accesses():
            extent = access.to_extent(horizontal_extent)
            if extent is None:
                continue

            if access.field in ctx.fields:
                ctx.fields[access.field] = ctx.fields[access.field] | extent
            else:
                ctx.fields[access.field] = extent


def compute_horizontal_block_extents(node: oir.Stencil, **kwargs: Any) -> Dict[int, Extent]:
    ctx = StencilExtentComputer(**kwargs).visit(node)
    return ctx.blocks


def compute_fields_extents(node: oir.Stencil, **kwargs: Any) -> Dict[str, Extent]:
    ctx = StencilExtentComputer(**kwargs).visit(node)
    return ctx.fields


def compute_extents(
    node: oir.Stencil, **kwargs: Any
) -> Tuple[Dict[str, Extent], Dict[int, Extent]]:
    ctx = StencilExtentComputer(**kwargs).visit(node)
    return ctx.fields, ctx.blocks



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Global configuration of Hypothesis and pytest for running tests."""


import hypothesis as hyp


# Ignore hidden folders and disabled tests
collect_ignore_glob = [".*", "_disabled*"]


def pytest_configure(config):
    # HealthCheck.too_slow causes more trouble than good -- especially in CIs.
    hyp.settings.register_profile(
        "slow", hyp.settings(suppress_health_check=[hyp.HealthCheck.too_slow], deadline=None)
    )
    config.addinivalue_line(
        "markers",
        "requires_gpu: mark tests that require a Nvidia GPU (assume cupy and cudatoolkit are installed)",
    )
    config.addinivalue_line(
        "markers",
        "requires_dace: mark tests that require dace in the python environment",
    )
    hyp.settings.load_profile("slow")


def pytest_addoption(parser):
    parser.addoption("--keep-gtcache", action="store_true", default=False, dest="keep_gtcache")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
from typing import Any, Dict, Optional, Type, Union

from gt4py.cartesian import utils as gt_utils
from gt4py.cartesian.definitions import BuildOptions, StencilID
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.type_hints import AnnotatedStencilFunc, StencilFunc


REGISTRY = gt_utils.Registry()
AnyStencilFunc = Union[StencilFunc, AnnotatedStencilFunc]


def from_name(name: str) -> Optional[Type["Frontend"]]:
    """Return frontend by name."""
    return REGISTRY.get(name, None)


def register(frontend_cls: Type["Frontend"]) -> None:
    """Register a new frontend."""
    return REGISTRY.register(frontend_cls.name, frontend_cls)


class Frontend(abc.ABC):

    name: str
    """Frontend name."""

    @classmethod
    @abc.abstractmethod
    def get_stencil_id(
        cls,
        qualified_name: str,
        definition: AnyStencilFunc,
        externals: Dict[str, Any],
        options_id: str,
    ) -> StencilID:
        """
        Create a StencilID object that contains a unique hash for the stencil.

        Notes
        -----
        This method seems to no longer be used through StencilBuilder.

        Returns
        -------
        StencilID:
            An object that contains the qualified name and unique hash.

        Raises
        ------
        GTSyntaxError
            If there is a parsing error.

        TypeError
            If there is a error resolving external types.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def generate(
        cls, definition: AnyStencilFunc, externals: Dict[str, Any], options: BuildOptions
    ) -> gtir.Stencil:
        """
        Generate a StencilDefinition from a stencil Python function.

        Raises
        ------
        GTScriptSyntaxError
            If there is a parsing error.

        GTScriptDefinitionError
            If types are misused in the definition.

        TypeError
            If there is an error in resolving external types.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def prepare_stencil_definition(
        cls, definition: AnyStencilFunc, externals: Dict[str, Any]
    ) -> AnnotatedStencilFunc:
        """
        Annotate the stencil function if not already done so.

        Raises
        ------
        GTSyntaxError
            If there is a parsing error.

        TypeError
            If there is a error resolving external types.
        """
        pass



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Union, cast

import dace
import dace.data
import dace.library
import dace.subsets

from gt4py import eve
from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc import daceir as dcir
from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.dace.expansion_specification import Loop, Map, Sections, Stages
from gt4py.cartesian.gtc.dace.utils import (
    compute_dcir_access_infos,
    flatten_list,
    get_tasklet_symbol,
    union_inout_memlets,
    union_node_grid_subsets,
    untile_memlets,
)
from gt4py.cartesian.gtc.definitions import Extent

from .utils import remove_horizontal_region


if TYPE_CHECKING:
    from gt4py.cartesian.gtc.dace.nodes import StencilComputation


def _access_iter(node: oir.HorizontalExecution, get_outputs: bool):
    if get_outputs:
        iterator = filter(
            lambda node: isinstance(node, oir.FieldAccess),
            node.walk_values().if_isinstance(oir.AssignStmt).getattr("left"),
        )
    else:

        def _iterator():
            for n in node.walk_values():
                if isinstance(n, oir.AssignStmt):
                    yield from n.right.walk_values().if_isinstance(oir.FieldAccess)
                elif isinstance(n, oir.While):
                    yield from n.cond.walk_values().if_isinstance(oir.FieldAccess)
                elif isinstance(n, oir.MaskStmt):
                    yield from n.mask.walk_values().if_isinstance(oir.FieldAccess)

        iterator = _iterator()

    yield from (
        eve.utils.xiter(iterator).map(
            lambda acc: (
                acc.name,
                acc.offset,
                get_tasklet_symbol(acc.name, acc.offset, is_target=get_outputs),
            )
        )
    ).unique(key=lambda x: x[2])


def _get_tasklet_inout_memlets(node: oir.HorizontalExecution, *, get_outputs, global_ctx, **kwargs):

    access_infos = compute_dcir_access_infos(
        node,
        block_extents=global_ctx.library_node.get_extents,
        oir_decls=global_ctx.library_node.declarations,
        collect_read=not get_outputs,
        collect_write=get_outputs,
        **kwargs,
    )

    res = list()
    for name, offset, tasklet_symbol in _access_iter(node, get_outputs=get_outputs):
        access_info = access_infos[name]
        if not access_info.variable_offset_axes:
            offset_dict = offset.to_dict()
            for axis in access_info.axes():
                access_info = access_info.restricted_to_index(
                    axis, extent=(offset_dict[axis.lower()], offset_dict[axis.lower()])
                )

        memlet = dcir.Memlet(
            field=name,
            connector=tasklet_symbol,
            access_info=access_info,
            is_read=not get_outputs,
            is_write=get_outputs,
        )
        res.append(memlet)
    return res


def _all_stmts_same_region(scope_nodes, axis: dcir.Axis, interval):
    def all_statements_in_region(scope_nodes):
        return all(
            isinstance(stmt, dcir.HorizontalRestriction)
            for tasklet in eve.walk_values(scope_nodes).if_isinstance(dcir.Tasklet)
            for stmt in tasklet.stmts
        )

    def all_regions_same(scope_nodes):
        return (
            len(
                set(
                    (
                        None
                        if mask.intervals[axis.to_idx()].start is None
                        else mask.intervals[axis.to_idx()].start.level,
                        None
                        if mask.intervals[axis.to_idx()].start is None
                        else mask.intervals[axis.to_idx()].start.offset,
                        None
                        if mask.intervals[axis.to_idx()].end is None
                        else mask.intervals[axis.to_idx()].end.level,
                        None
                        if mask.intervals[axis.to_idx()].end is None
                        else mask.intervals[axis.to_idx()].end.offset,
                    )
                    for mask in eve.walk_values(scope_nodes).if_isinstance(common.HorizontalMask)
                )
            )
            == 1
        )

    return (
        axis in dcir.Axis.dims_horizontal()
        and isinstance(interval, dcir.DomainInterval)
        and all_statements_in_region(scope_nodes)
        and all_regions_same(scope_nodes)
    )


class DaCeIRBuilder(eve.NodeTranslator):
    @dataclass
    class GlobalContext:
        library_node: "StencilComputation"
        arrays: Dict[str, dace.data.Data]

        def get_dcir_decls(
            self,
            access_infos: Dict[eve.SymbolRef, dcir.FieldAccessInfo],
            symbol_collector: "DaCeIRBuilder.SymbolCollector",
        ) -> List[dcir.FieldDecl]:
            return [
                self._get_dcir_decl(field, access_info, symbol_collector=symbol_collector)
                for field, access_info in access_infos.items()
            ]

        def _get_dcir_decl(
            self,
            field: eve.SymbolRef,
            access_info: dcir.FieldAccessInfo,
            symbol_collector: "DaCeIRBuilder.SymbolCollector",
        ) -> dcir.FieldDecl:
            oir_decl: oir.Decl = self.library_node.declarations[field]
            assert isinstance(oir_decl, oir.FieldDecl)
            dace_array = self.arrays[field]
            for s in dace_array.strides:
                for sym in dace.symbolic.symlist(s).values():
                    symbol_collector.add_symbol(str(sym))
            for sym in access_info.grid_subset.free_symbols:
                symbol_collector.add_symbol(sym)

            return dcir.FieldDecl(
                name=field,
                dtype=oir_decl.dtype,
                strides=tuple(str(s) for s in dace_array.strides),
                data_dims=oir_decl.data_dims,
                access_info=access_info,
                storage=dcir.StorageType.from_dace_storage(dace.StorageType.Default),
            )

    @dataclass
    class IterationContext:
        grid_subset: dcir.GridSubset
        parent: Optional["DaCeIRBuilder.IterationContext"]

        @classmethod
        def init(cls, *args, **kwargs):
            res = cls(*args, parent=None, **kwargs)
            return res

        def push_axes_extents(self, axes_extents) -> "DaCeIRBuilder.IterationContext":
            res = self.grid_subset
            for axis, extent in axes_extents.items():
                axis_interval = res.intervals[axis]
                if isinstance(axis_interval, dcir.DomainInterval):
                    res__interval = dcir.DomainInterval(
                        start=dcir.AxisBound(
                            level=common.LevelMarker.START, offset=extent[0], axis=axis
                        ),
                        end=dcir.AxisBound(
                            level=common.LevelMarker.END, offset=extent[1], axis=axis
                        ),
                    )
                    res = res.set_interval(axis, res__interval)
                elif isinstance(axis_interval, dcir.TileInterval):
                    tile_interval = dcir.TileInterval(
                        axis=axis,
                        start_offset=extent[0],
                        end_offset=extent[1],
                        tile_size=axis_interval.tile_size,
                        domain_limit=axis_interval.domain_limit,
                    )
                    res = res.set_interval(axis, tile_interval)
                # if is IndexWithExtent, do nothing.
            return DaCeIRBuilder.IterationContext(grid_subset=res, parent=self)

        def push_interval(
            self, axis: dcir.Axis, interval: Union[dcir.DomainInterval, oir.Interval]
        ) -> "DaCeIRBuilder.IterationContext":
            return DaCeIRBuilder.IterationContext(
                grid_subset=self.grid_subset.set_interval(axis, interval),
                parent=self,
            )

        def push_expansion_item(self, item: Union[Map, Loop]) -> "DaCeIRBuilder.IterationContext":

            if not isinstance(item, (Map, Loop)):
                raise ValueError

            if isinstance(item, Map):
                iterations = item.iterations
            else:
                iterations = [item]

            grid_subset = self.grid_subset
            for it in iterations:
                axis = it.axis
                if it.kind == "tiling":
                    assert it.stride is not None
                    grid_subset = grid_subset.tile(tile_sizes={axis: it.stride})
                else:
                    grid_subset = grid_subset.restricted_to_index(axis)
            return DaCeIRBuilder.IterationContext(grid_subset=grid_subset, parent=self)

        def push_expansion_items(
            self, items: Iterable[Union[Map, Loop]]
        ) -> "DaCeIRBuilder.IterationContext":
            res = self
            for item in items:
                res = res.push_expansion_item(item)
            return res

        def pop(self) -> "DaCeIRBuilder.IterationContext":
            assert self.parent is not None
            return self.parent

    @dataclass
    class SymbolCollector:
        symbol_decls: Dict[str, dcir.SymbolDecl] = dataclasses.field(default_factory=dict)

        def add_symbol(self, name: str, dtype: common.DataType = common.DataType.INT32):
            if name not in self.symbol_decls:
                self.symbol_decls[name] = dcir.SymbolDecl(name=name, dtype=dtype)
            else:
                assert self.symbol_decls[name].dtype == dtype

        def remove_symbol(self, name: eve.SymbolRef):
            if name in self.symbol_decls:
                del self.symbol_decls[name]

    def visit_Literal(self, node: oir.Literal, **kwargs: Any) -> dcir.Literal:
        return dcir.Literal(value=node.value, dtype=node.dtype)

    def visit_UnaryOp(self, node: oir.UnaryOp, **kwargs: Any) -> dcir.UnaryOp:
        return dcir.UnaryOp(op=node.op, expr=self.visit(node.expr, **kwargs), dtype=node.dtype)

    def visit_BinaryOp(self, node: oir.BinaryOp, **kwargs: Any) -> dcir.BinaryOp:
        return dcir.BinaryOp(
            op=node.op,
            left=self.visit(node.left, **kwargs),
            right=self.visit(node.right, **kwargs),
            dtype=node.dtype,
        )

    def visit_HorizontalRestriction(
        self,
        node: oir.HorizontalRestriction,
        *,
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
        **kwargs: Any,
    ) -> dcir.HorizontalRestriction:
        for axis, interval in zip(dcir.Axis.dims_horizontal(), node.mask.intervals):
            for bound in (interval.start, interval.end):
                if bound is not None:
                    symbol_collector.add_symbol(axis.iteration_symbol())
                    if bound.level == common.LevelMarker.END:
                        symbol_collector.add_symbol(axis.domain_symbol())
        return dcir.HorizontalRestriction(
            mask=node.mask,
            body=self.visit(node.body, symbol_collector=symbol_collector, **kwargs),
        )

    def visit_VariableKOffset(self, node: oir.VariableKOffset, **kwargs):
        return dcir.VariableKOffset(k=self.visit(node.k, **kwargs))

    def visit_LocalScalar(self, node: oir.LocalScalar, **kwargs: Any) -> dcir.LocalScalarDecl:
        return dcir.LocalScalarDecl(name=node.name, dtype=node.dtype)

    def visit_FieldAccess(
        self,
        node: oir.FieldAccess,
        *,
        is_target: bool,
        targets: Set[eve.SymbolRef],
        var_offset_fields: Set[eve.SymbolRef],
        **kwargs: Any,
    ) -> Union[dcir.IndexAccess, dcir.ScalarAccess]:
        res: Union[dcir.IndexAccess, dcir.ScalarAccess]
        if node.name in var_offset_fields:
            res = dcir.IndexAccess(
                name=node.name + "__",
                offset=self.visit(
                    node.offset,
                    is_target=False,
                    targets=targets,
                    var_offset_fields=var_offset_fields,
                    **kwargs,
                ),
                data_index=node.data_index,
                dtype=node.dtype,
            )
        else:
            is_target = is_target or (
                node.name in targets and node.offset == common.CartesianOffset.zero()
            )
            name = get_tasklet_symbol(node.name, node.offset, is_target=is_target)
            if node.data_index:
                res = dcir.IndexAccess(
                    name=name, offset=None, data_index=node.data_index, dtype=node.dtype
                )
            else:
                res = dcir.ScalarAccess(name=name, dtype=node.dtype)
        if is_target:
            targets.add(node.name)
        return res

    def visit_ScalarAccess(
        self,
        node: oir.ScalarAccess,
        *,
        global_ctx: "DaCeIRBuilder.GlobalContext",
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
        **kwargs: Any,
    ) -> dcir.ScalarAccess:
        if node.name in global_ctx.library_node.declarations:
            symbol_collector.add_symbol(node.name, dtype=node.dtype)
        return dcir.ScalarAccess(name=node.name, dtype=node.dtype)

    def visit_AssignStmt(self, node: oir.AssignStmt, *, targets, **kwargs: Any) -> dcir.AssignStmt:
        # the visiting order matters here, since targets must not contain the target symbols from the left visit
        right = self.visit(node.right, is_target=False, targets=targets, **kwargs)
        left = self.visit(node.left, is_target=True, targets=targets, **kwargs)
        return dcir.AssignStmt(left=left, right=right)

    def visit_MaskStmt(self, node: oir.MaskStmt, **kwargs: Any) -> dcir.MaskStmt:
        return dcir.MaskStmt(
            mask=self.visit(node.mask, is_target=False, **kwargs),
            body=self.visit(node.body, **kwargs),
        )

    def visit_While(self, node: oir.While, **kwargs: Any) -> dcir.While:
        return dcir.While(
            cond=self.visit(node.cond, is_target=False, **kwargs),
            body=self.visit(node.body, **kwargs),
        )

    def visit_Cast(self, node: oir.Cast, **kwargs: Any) -> dcir.Cast:
        return dcir.Cast(dtype=node.dtype, expr=self.visit(node.expr, **kwargs))

    def visit_NativeFuncCall(self, node: oir.NativeFuncCall, **kwargs: Any) -> dcir.NativeFuncCall:
        return dcir.NativeFuncCall(
            func=node.func, args=self.visit(node.args, **kwargs), dtype=node.dtype
        )

    def visit_TernaryOp(self, node: oir.TernaryOp, **kwargs: Any) -> dcir.TernaryOp:
        return dcir.TernaryOp(
            cond=self.visit(node.cond, **kwargs),
            true_expr=self.visit(node.true_expr, **kwargs),
            false_expr=self.visit(node.false_expr, **kwargs),
            dtype=node.dtype,
        )

    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        *,
        global_ctx: "DaCeIRBuilder.GlobalContext",
        iteration_ctx: "DaCeIRBuilder.IterationContext",
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
        loop_order,
        k_interval,
        **kwargs,
    ):
        # skip type checking due to https://github.com/python/mypy/issues/5485
        extent = global_ctx.library_node.get_extents(node)  # type: ignore
        decls = [self.visit(decl, **kwargs) for decl in node.declarations]
        targets: Set[str] = set()
        stmts = [
            self.visit(
                stmt,
                targets=targets,
                global_ctx=global_ctx,
                symbol_collector=symbol_collector,
                **kwargs,
            )
            for stmt in node.body
        ]

        stages_idx = next(
            idx
            for idx, item in enumerate(global_ctx.library_node.expansion_specification)
            if isinstance(item, Stages)
        )
        expansion_items = global_ctx.library_node.expansion_specification[stages_idx + 1 :]

        iteration_ctx = iteration_ctx.push_axes_extents(
            {k: v for k, v in zip(dcir.Axis.dims_horizontal(), extent)}
        )
        iteration_ctx = iteration_ctx.push_expansion_items(expansion_items)

        assert iteration_ctx.grid_subset == dcir.GridSubset.single_gridpoint()

        read_memlets = _get_tasklet_inout_memlets(
            node,
            get_outputs=False,
            global_ctx=global_ctx,
            grid_subset=iteration_ctx.grid_subset,
            k_interval=k_interval,
        )

        write_memlets = _get_tasklet_inout_memlets(
            node,
            get_outputs=True,
            global_ctx=global_ctx,
            grid_subset=iteration_ctx.grid_subset,
            k_interval=k_interval,
        )

        dcir_node = dcir.Tasklet(
            decls=decls,
            stmts=stmts,
            read_memlets=read_memlets,
            write_memlets=write_memlets,
        )

        for item in reversed(expansion_items):
            iteration_ctx = iteration_ctx.pop()
            dcir_node = self._process_iteration_item(
                [dcir_node],
                item,
                global_ctx=global_ctx,
                iteration_ctx=iteration_ctx,
                symbol_collector=symbol_collector,
                **kwargs,
            )
        # pop stages context (pushed with push_grid_subset)
        iteration_ctx.pop()
        return dcir_node

    def visit_VerticalLoopSection(
        self,
        node: oir.VerticalLoopSection,
        *,
        loop_order,
        iteration_ctx: "DaCeIRBuilder.IterationContext",
        global_ctx: "DaCeIRBuilder.GlobalContext",
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
        **kwargs,
    ):
        sections_idx, stages_idx = [
            idx
            for idx, item in enumerate(global_ctx.library_node.expansion_specification)
            if isinstance(item, (Sections, Stages))
        ]
        expansion_items = global_ctx.library_node.expansion_specification[
            sections_idx + 1 : stages_idx
        ]

        iteration_ctx = iteration_ctx.push_interval(
            dcir.Axis.K, node.interval
        ).push_expansion_items(expansion_items)

        dcir_nodes = self.generic_visit(
            node.horizontal_executions,
            iteration_ctx=iteration_ctx,
            global_ctx=global_ctx,
            symbol_collector=symbol_collector,
            loop_order=loop_order,
            k_interval=node.interval,
            **kwargs,
        )

        # if multiple horizontal executions, enforce their order by means of a state machine
        if len(dcir_nodes) > 1:
            dcir_nodes = [
                self.to_state([node], grid_subset=node.grid_subset)
                for node in flatten_list(dcir_nodes)
            ]

        for item in reversed(expansion_items):
            iteration_ctx = iteration_ctx.pop()
            dcir_nodes = self._process_iteration_item(
                scope=dcir_nodes,
                item=item,
                iteration_ctx=iteration_ctx,
                global_ctx=global_ctx,
                symbol_collector=symbol_collector,
            )
        # pop off interval
        iteration_ctx.pop()
        return dcir_nodes

    def to_dataflow(
        self,
        nodes,
        *,
        global_ctx: "DaCeIRBuilder.GlobalContext",
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
    ):

        nodes = flatten_list(nodes)
        if all(isinstance(n, (dcir.NestedSDFG, dcir.DomainMap, dcir.Tasklet)) for n in nodes):
            return nodes
        elif not all(isinstance(n, (dcir.ComputationState, dcir.DomainLoop)) for n in nodes):
            raise ValueError("Can't mix dataflow and state nodes on same level.")

        read_memlets, write_memlets, field_memlets = union_inout_memlets(nodes)

        field_decls = global_ctx.get_dcir_decls(
            {memlet.field: memlet.access_info for memlet in field_memlets},
            symbol_collector=symbol_collector,
        )
        read_fields = {memlet.field for memlet in read_memlets}
        write_fields = {memlet.field for memlet in write_memlets}
        read_memlets = [
            memlet.remove_write() for memlet in field_memlets if memlet.field in read_fields
        ]
        write_memlets = [
            memlet.remove_read() for memlet in field_memlets if memlet.field in write_fields
        ]
        return [
            dcir.NestedSDFG(
                label=global_ctx.library_node.label,
                field_decls=field_decls,
                # NestedSDFG must have same shape on input and output, matching corresponding
                # nsdfg.sdfg's array shape
                read_memlets=read_memlets,
                write_memlets=write_memlets,
                states=nodes,
                symbol_decls=list(symbol_collector.symbol_decls.values()),
            )
        ]

    def to_state(self, nodes, *, grid_subset: dcir.GridSubset):

        nodes = flatten_list(nodes)
        if all(isinstance(n, (dcir.ComputationState, dcir.DomainLoop)) for n in nodes):
            return nodes
        elif all(isinstance(n, (dcir.NestedSDFG, dcir.DomainMap, dcir.Tasklet)) for n in nodes):
            return [dcir.ComputationState(computations=nodes, grid_subset=grid_subset)]
        else:
            raise ValueError("Can't mix dataflow and state nodes on same level.")

    def _process_map_item(
        self,
        scope_nodes,
        item: Map,
        *,
        global_ctx,
        iteration_ctx: "DaCeIRBuilder.IterationContext",
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
        **kwargs,
    ):

        grid_subset = iteration_ctx.grid_subset
        read_memlets, write_memlets, _ = union_inout_memlets(list(scope_nodes))
        scope_nodes = self.to_dataflow(
            scope_nodes, global_ctx=global_ctx, symbol_collector=symbol_collector
        )

        ranges = []
        for iteration in item.iterations:
            axis = iteration.axis
            interval = iteration_ctx.grid_subset.intervals[axis]
            grid_subset = grid_subset.set_interval(axis, interval)
            if iteration.kind == "tiling":
                read_memlets = untile_memlets(read_memlets, axes=[axis])
                write_memlets = untile_memlets(write_memlets, axes=[axis])
                if not axis == dcir.Axis.K:
                    interval = dcir.DomainInterval(
                        start=dcir.AxisBound.from_common(axis, oir.AxisBound.start()),
                        end=dcir.AxisBound.from_common(axis, oir.AxisBound.end()),
                    )
                symbol_collector.remove_symbol(axis.tile_symbol())
                ranges.append(
                    dcir.Range(
                        var=axis.tile_symbol(),
                        interval=interval,
                        stride=iteration.stride,
                    )
                )
            else:
                if _all_stmts_same_region(scope_nodes, axis, interval):
                    masks = cast(
                        List[common.HorizontalMask],
                        eve.walk_values(scope_nodes).if_isinstance(common.HorizontalMask).to_list(),
                    )
                    horizontal_mask_interval = next(
                        iter((mask.intervals[axis.to_idx()] for mask in masks))
                    )
                    interval = dcir.DomainInterval.intersection(
                        axis, horizontal_mask_interval, interval
                    )
                    scope_nodes = remove_horizontal_region(scope_nodes, axis)
                assert iteration.kind == "contiguous"
                res_read_memlets = []
                res_write_memlets = []
                for memlet in read_memlets:
                    access_info = memlet.access_info.apply_iteration(
                        dcir.GridSubset.from_interval(interval, axis)
                    )
                    for sym in access_info.grid_subset.free_symbols:
                        symbol_collector.add_symbol(sym)
                    res_read_memlets.append(
                        dcir.Memlet(
                            field=memlet.field,
                            connector=memlet.connector,
                            access_info=access_info,
                            is_read=True,
                            is_write=False,
                        )
                    )
                for memlet in write_memlets:
                    access_info = memlet.access_info.apply_iteration(
                        dcir.GridSubset.from_interval(interval, axis)
                    )
                    for sym in access_info.grid_subset.free_symbols:
                        symbol_collector.add_symbol(sym)
                    res_write_memlets.append(
                        dcir.Memlet(
                            field=memlet.field,
                            connector=memlet.connector,
                            access_info=access_info,
                            is_read=False,
                            is_write=True,
                        )
                    )
                read_memlets = res_read_memlets
                write_memlets = res_write_memlets

                assert not isinstance(interval, dcir.IndexWithExtent)
                index_range = dcir.Range.from_axis_and_interval(axis, interval)
                symbol_collector.remove_symbol(index_range.var)
                ranges.append(index_range)

        return [
            dcir.DomainMap(
                computations=scope_nodes,
                index_ranges=ranges,
                schedule=dcir.MapSchedule.from_dace_schedule(item.schedule),
                read_memlets=read_memlets,
                write_memlets=write_memlets,
                grid_subset=grid_subset,
            )
        ]

    def _process_loop_item(
        self,
        scope_nodes,
        item: Loop,
        *,
        global_ctx,
        iteration_ctx: "DaCeIRBuilder.IterationContext",
        symbol_collector: "DaCeIRBuilder.SymbolCollector",
        **kwargs,
    ):

        grid_subset = union_node_grid_subsets(list(scope_nodes))
        read_memlets, write_memlets, _ = union_inout_memlets(list(scope_nodes))
        scope_nodes = self.to_state(scope_nodes, grid_subset=grid_subset)

        axis = item.axis
        interval = iteration_ctx.grid_subset.intervals[axis]
        grid_subset = grid_subset.set_interval(axis, interval)
        if item.kind == "tiling":
            raise NotImplementedError("Tiling as a state machine not implemented.")

        assert item.kind == "contiguous"
        res_read_memlets = []
        res_write_memlets = []
        for memlet in read_memlets:
            access_info = memlet.access_info.apply_iteration(
                dcir.GridSubset.from_interval(interval, axis)
            )
            for sym in access_info.grid_subset.free_symbols:
                symbol_collector.add_symbol(sym)
            res_read_memlets.append(
                dcir.Memlet(
                    field=memlet.field,
                    connector=memlet.connector,
                    access_info=access_info,
                    is_read=True,
                    is_write=False,
                )
            )
        for memlet in write_memlets:
            access_info = memlet.access_info.apply_iteration(
                dcir.GridSubset.from_interval(interval, axis)
            )
            for sym in access_info.grid_subset.free_symbols:
                symbol_collector.add_symbol(sym)
            res_write_memlets.append(
                dcir.Memlet(
                    field=memlet.field,
                    connector=memlet.connector,
                    access_info=access_info,
                    is_read=False,
                    is_write=True,
                )
            )
        read_memlets = res_read_memlets
        write_memlets = res_write_memlets

        assert not isinstance(interval, dcir.IndexWithExtent)
        index_range = dcir.Range.from_axis_and_interval(axis, interval, stride=item.stride)
        for sym in index_range.free_symbols:
            symbol_collector.add_symbol(sym, common.DataType.INT32)
        symbol_collector.remove_symbol(index_range.var)
        return [
            dcir.DomainLoop(
                axis=axis,
                loop_states=scope_nodes,
                index_range=index_range,
                read_memlets=read_memlets,
                write_memlets=write_memlets,
                grid_subset=grid_subset,
            )
        ]

    def _process_iteration_item(self, scope, item, **kwargs):
        if isinstance(item, Map):
            return self._process_map_item(scope, item, **kwargs)
        elif isinstance(item, Loop):
            return self._process_loop_item(scope, item, **kwargs)
        else:
            raise ValueError("Invalid expansion specification set.")

    def visit_VerticalLoop(
        self,
        node: oir.VerticalLoop,
        *,
        global_ctx: "DaCeIRBuilder.GlobalContext",
        **kwargs,
    ):
        start, end = (node.sections[0].interval.start, node.sections[0].interval.end)

        overall_interval = dcir.DomainInterval(
            start=dcir.AxisBound(axis=dcir.Axis.K, level=start.level, offset=start.offset),
            end=dcir.AxisBound(axis=dcir.Axis.K, level=end.level, offset=end.offset),
        )
        overall_extent = Extent.zeros(2)
        for he in node.walk_values().if_isinstance(oir.HorizontalExecution):
            overall_extent = overall_extent.union(global_ctx.library_node.get_extents(he))

        iteration_ctx = DaCeIRBuilder.IterationContext.init(
            grid_subset=dcir.GridSubset.from_gt4py_extent(overall_extent).set_interval(
                axis=dcir.Axis.K, interval=overall_interval
            )
        )

        var_offset_fields = {
            acc.name
            for acc in node.walk_values().if_isinstance(oir.FieldAccess)
            if isinstance(acc.offset, oir.VariableKOffset)
        }
        sections_idx = next(
            idx
            for idx, item in enumerate(global_ctx.library_node.expansion_specification)
            if isinstance(item, Sections)
        )
        expansion_items = global_ctx.library_node.expansion_specification[:sections_idx]
        iteration_ctx = iteration_ctx.push_expansion_items(expansion_items)

        symbol_collector = DaCeIRBuilder.SymbolCollector()
        sections = flatten_list(
            self.generic_visit(
                node.sections,
                loop_order=node.loop_order,
                global_ctx=global_ctx,
                iteration_ctx=iteration_ctx,
                symbol_collector=symbol_collector,
                var_offset_fields=var_offset_fields,
                **kwargs,
            )
        )
        if node.loop_order != common.LoopOrder.PARALLEL:
            sections = [self.to_state(s, grid_subset=iteration_ctx.grid_subset) for s in sections]
        computations = sections
        for item in reversed(expansion_items):
            iteration_ctx = iteration_ctx.pop()
            computations = self._process_iteration_item(
                scope=computations,
                item=item,
                iteration_ctx=iteration_ctx,
                global_ctx=global_ctx,
                symbol_collector=symbol_collector,
            )

        read_memlets, write_memlets, field_memlets = union_inout_memlets(computations)

        field_decls = global_ctx.get_dcir_decls(
            global_ctx.library_node.access_infos,
            symbol_collector=symbol_collector,
        )

        read_fields = set(memlet.field for memlet in read_memlets)
        write_fields = set(memlet.field for memlet in write_memlets)
        res = dcir.NestedSDFG(
            label=global_ctx.library_node.label,
            states=self.to_state(computations, grid_subset=iteration_ctx.grid_subset),
            field_decls=field_decls,
            read_memlets=[memlet for memlet in field_memlets if memlet.field in read_fields],
            write_memlets=[memlet for memlet in field_memlets if memlet.field in write_fields],
            symbol_decls=list(symbol_collector.symbol_decls.values()),
        )

        return res



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gt4py.cartesian.gtc import gtir


def prune_unused_parameters(node: gtir.Stencil) -> gtir.Stencil:
    """
    Remove unused parameters from the gtir signature.

    (Maybe this pass should go into a later stage. If you need to touch this pass,
    e.g. when the definition_ir gets removed, consider moving it to a more appropriate
    level. Maybe to the backend IR?)
    """
    assert isinstance(node, gtir.Stencil)
    used_variables = (
        node.walk_values()
        .if_isinstance(gtir.FieldAccess, gtir.ScalarAccess)
        .getattr("name")
        .to_list()
    )
    used_params = list(filter(lambda param: param.name in used_variables, node.params))
    return node.copy(update={"params": used_params})



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Set, Tuple

import dace
from dace.frontend.python.common import SDFGConvertible

from gt4py.cartesian.backend.dace_backend import SDFGManager
from gt4py.cartesian.backend.dace_stencil_object import DaCeStencilObject, add_optional_fields
from gt4py.cartesian.backend.module_generator import make_args_data_from_gtir
from gt4py.cartesian.lazy_stencil import LazyStencil


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_builder import StencilBuilder


class DaCeLazyStencil(LazyStencil, SDFGConvertible):
    def __init__(self, builder: "StencilBuilder"):
        if "dace" not in builder.backend.name:
            raise ValueError("Trying to build a DaCeLazyStencil for non-dace backend.")
        super().__init__(builder=builder)

    @property
    def field_info(self) -> Dict[str, Any]:
        """
        Return same value as compiled stencil object's `field_info` attribute.

        Does not trigger a build.
        """
        return make_args_data_from_gtir(self.builder.gtir_pipeline).field_info

    def closure_resolver(
        self,
        constant_args: Dict[str, Any],
        given_args: Set[str],
        parent_closure: Optional["dace.frontend.python.common.SDFGClosure"] = None,
    ) -> "dace.frontend.python.common.SDFGClosure":
        return dace.frontend.python.common.SDFGClosure()

    def __sdfg__(self, *args, **kwargs) -> dace.SDFG:
        sdfg_manager = SDFGManager(self.builder)
        args_data = make_args_data_from_gtir(self.builder.gtir_pipeline)
        arg_names = [arg.name for arg in self.builder.gtir.api_signature]
        assert args_data.domain_info is not None
        norm_kwargs = DaCeStencilObject.normalize_args(
            *args,
            backend=self.backend.name,
            arg_names=arg_names,
            domain_info=args_data.domain_info,
            field_info=args_data.field_info,
            **kwargs,
        )
        sdfg = sdfg_manager.frozen_sdfg(origin=norm_kwargs["origin"], domain=norm_kwargs["domain"])
        return add_optional_fields(
            sdfg,
            field_info=args_data.field_info,
            parameter_info=args_data.parameter_info,
            **norm_kwargs,
        )

    def __sdfg_closure__(self, reevaluate: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
        return {}

    def __sdfg_signature__(self) -> Tuple[Sequence[str], Sequence[str]]:
        args = [arg.name for arg in self.builder.gtir.api_signature]
        return (args, [])



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Implementation of GTScript: an embedded DSL in Python for stencil computations.

Interface functions to define and compile GTScript definitions and empty symbol
definitions for the keywords of the DSL.
"""

import sys

from gt4py.cartesian.gtscript import *


sys.modules["__gtscript__"] = sys.modules["gt4py.__gtscript__"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Definitions of node and visitor trait classes."""


from __future__ import annotations

import collections

from . import concepts, datamodels, exceptions, visitors
from .extended_typing import Any, Dict, Set, Type, no_type_check


# ---  Node Traits ---
@concepts.register_annex_user("symtable", Dict[str, concepts.Node], shared=True)
@datamodels.datamodel
class SymbolTableTrait:
    """Node trait adding an automatically created symbol table to the parent node.

    The actual symbol table dict will be stored in the `annex.symtable` attribute.
    To inject extra symbol definitions, add the nodes to a class attribute
    called ``_NODE_SYMBOLS_``.
    """

    __slots__ = ()

    @no_type_check
    @datamodels.root_validator
    @classmethod
    def _collect_symbol_names(cls: Type[SymbolTableTrait], instance: concepts.Node) -> None:
        collected_symbols = cls.SymbolsCollector.apply(instance)
        instance.annex.symtable = collected_symbols

    class SymbolsCollector(visitors.NodeVisitor):
        def __init__(self) -> None:
            self.collected_symbols: Dict[str, concepts.Node] = {}

        def visit_Node(self, node: concepts.Node, /) -> None:
            for field_name, attribute in node.__datamodel_fields__.items():
                if isinstance(attribute.type, type) and issubclass(
                    attribute.type, concepts.SymbolName
                ):
                    symbol_name = getattr(node, field_name)
                    if symbol_name in self.collected_symbols:
                        raise exceptions.EveValueError(
                            f"Multiple definitions of symbol '{symbol_name}'"
                        )
                    self.collected_symbols[symbol_name] = node

            if not isinstance(node, SymbolTableTrait):
                # Stop recursion if the node opens a new scope (i.e. node with SymbolTableTrait)
                self.generic_visit(node)

        def visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
            if hasattr(node.__class__, "_NODE_SYMBOLS_"):
                self.visit(node.__class__._NODE_SYMBOLS_)  # type: ignore[union-attr]  # _NODE_SYMBOLS_ is optional
            return super().visit(node, **kwargs)

        @classmethod
        def apply(cls, node: concepts.Node) -> Dict[str, concepts.Node]:
            collector = cls()
            # If the passed root node already contains a symbol table, the check in `visit_Node()`
            # will automatically stop the traversal. To avoid this premature stop, we start the
            # traversal here calling `generic_visit()` to directly inspect the children (after
            # adding any extra node symbols defined in the node class).
            if hasattr(node.__class__, "_NODE_SYMBOLS_"):
                collector.visit(node.__class__._NODE_SYMBOLS_)  # type: ignore[attr-defined]  # _NODE_SYMBOLS_ is optional
            collector.generic_visit(node)
            return collector.collected_symbols


@concepts.register_annex_user("symtable", Dict[str, concepts.Node], shared=True)
@datamodels.datamodel
class SymbolRefsValidatorTrait:
    """Node trait adding automatic validation of symbol references appearing the node tree.

    It assumes that the symbol table with the actual definitions is stored as
    a dict in the `annex.symtable` attribute (like :class:`SymbolTableTrait` does).
    """

    __slots__ = ()

    @no_type_check
    @datamodels.root_validator
    @classmethod
    def _validate_symbol_refs(cls: Type[SymbolRefsValidatorTrait], instance: concepts.Node) -> None:
        validator = cls.SymbolRefsValidator()
        symtable = instance.annex.symtable
        for child_node in instance.iter_children_values():
            validator.visit(child_node, symtable=symtable)

        if validator.missing_symbols:
            raise exceptions.EveValueError(
                "Symbols {} not found.".format(validator.missing_symbols)
            )

    class SymbolRefsValidator(visitors.NodeVisitor):
        def __init__(self) -> None:
            self.missing_symbols: Set[str] = set()

        def visit_Node(
            self, node: concepts.Node, *, symtable: Dict[str, Any], **kwargs: Any
        ) -> None:
            for field_name, attribute in node.__datamodel_fields__.items():
                if isinstance(attribute.type, type) and issubclass(
                    attribute.type, concepts.SymbolRef
                ):
                    symbol_name = getattr(node, field_name)
                    if symbol_name not in symtable:
                        self.missing_symbols.add(symbol_name)

            if isinstance(node, SymbolTableTrait):
                # Append symbols from nested scope for nested nodes
                symtable = {**symtable, **node.annex.symtable}
            self.generic_visit(node, symtable=symtable, **kwargs)

        @classmethod
        def apply(cls, node: concepts.Node, *, symtable: Dict[str, Any]) -> Set[str]:
            validator = cls()
            validator.visit(node, symtable=symtable)
            return validator.missing_symbols


class ValidatedSymbolTableTrait(SymbolRefsValidatorTrait, SymbolTableTrait):
    """Node trait adding an automatically created and validated symbol table.

    It is just the combination of the :class:`SymbolTableTrait` and
    :class:`SymbolRefsValidatorTrait` traits.
    """

    __slots__ = ()
    pass


# --- Visitor Traits ---
class VisitorWithSymbolTableTrait(visitors.NodeVisitor):
    """Visitor trait to update or add the symtable to kwargs in the visitor calls.

    Visitors inhering from this trait will automatically pass the active
    symbol table to visitor methods as the 'symtable' keyword argument.
    """

    def visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
        kwargs.setdefault("symtable", collections.ChainMap())
        new_scope = False
        if isinstance(node, concepts.Node):
            if new_scope := ("symtable" in node.annex):
                kwargs["symtable"] = kwargs["symtable"].new_child(node.annex.symtable)

        result = super().visit(node, **kwargs)

        if new_scope:
            kwargs["symtable"] = kwargs["symtable"].parents

        return result



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Stencil Object that allows for deferred building."""
from typing import TYPE_CHECKING, Any, Dict

from cached_property import cached_property


if TYPE_CHECKING:
    from gt4py.cartesian.backend.base import Backend
    from gt4py.cartesian.stencil_builder import StencilBuilder
    from gt4py.cartesian.stencil_object import StencilObject


class LazyStencil:
    """
    A stencil object which defers compilation until it is needed.

    Usually obtained using the :py:func:`gt4py.gtscript.lazy_stencil` decorator, not directly
    instanciated.
    This is done by keeping a reference to a :py:class:`gt4py.stencil_builder.StencilBuilder`
    instance.

    Compilation happens implicitly on first access to the `implementation` property.
    Low-level build utilities are accessible through the public :code:`builder` attribute.
    """

    def __init__(self, builder: "StencilBuilder"):
        self.builder = builder
        self.builder.caching.capture_externals()

    @cached_property
    def implementation(self) -> "StencilObject":
        """
        Expose the compiled backend-specific python callable which executes the stencil.

        Compilation happens at first access, the result is cached and should consecutively be
        accessible without overhead (not rigorously tested / benchmarked).
        """
        impl = self.builder.build()()
        return impl

    @property
    def backend(self) -> "Backend":
        """Do not trigger a build."""
        return self.builder.backend

    @property
    def field_info(self) -> Dict[str, Any]:
        """
        Access the compiled stencil object's `field_info` attribute.

        Triggers a build if necessary.
        """
        return self.implementation.field_info

    def check_syntax(self) -> None:
        """Create the gtscript IR for the stencil, failing on syntax errors."""
        if not self.builder.gtir:
            raise RuntimeError("Frontend did not raise a syntax error but did not generate IR.")

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        """Execute the stencil, building the stencil if necessary."""
        self.implementation(*args, **kwargs)

    def run(self, *args: Any, **kwargs: Any) -> None:
        """Pass through to the implementation.run."""
        self.implementation.run(*args, **kwargs)

    def __sdfg__(self, **kwargs):
        raise TypeError(
            f'Only dace backends are supported in DaCe-orchestrated programs. (found "{self.backend.name}")'
        )

    def __sdfg_signature__(self):
        raise TypeError(
            f'Only dace backends are supported in DaCe-orchestrated programs. (found "{self.backend.name}")'
        )

    def __sdfg_closure__(self, *args, **kwargs):
        raise TypeError(
            f'Only dace backends are supported in DaCe-orchestrated programs. (found "{self.backend.name}")'
        )
